from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd


ArrayLike = Union[np.ndarray, Sequence[float]]


@dataclass
class BlockJackknifeResult:
    est: float
    se: float
    z: float
    p: float
    n_blocks: int
    n_snps: int


def _compute_block_indices(
    n_snps: int,
    size: int = 5000,
    block_labels: Optional[np.ndarray] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Build SNP-to-block assignments using fixed SNP counts per block.

    Returns:
        block_ids (int array of shape (n_snps,)): Block index for each SNP (0..B-1)
        block_lengths (int array of shape (n_blocks,)): Number of SNPs per block
    """
    if block_labels is not None:
        block_labels = np.asarray(block_labels)
        if block_labels.shape[0] != n_snps:
            raise ValueError("'block_labels' must have length n_snps")
        # Reindex to 0..B-1 preserving first appearance order
        codes, uniques = pd.factorize(block_labels, sort=False)
        counts = np.bincount(codes, minlength=len(uniques))
        return codes.astype(np.int32), counts.astype(np.int32)

    size = max(1, int(size))
    block_ids = np.arange(n_snps) // size
    # block lengths: last block may be shorter
    n_blocks = int(block_ids.max()) + 1
    block_lengths = np.bincount(block_ids, minlength=n_blocks)
    return block_ids.astype(np.int32), block_lengths.astype(np.int32)


def _jackknife_ratio_from_block_sums(
    num_block_sums: np.ndarray,
    den_block_sums: np.ndarray,
) -> BlockJackknifeResult:
    """
    Compute statistic = sum(num) / sum(den) and its block-jackknife SE using delete-one blocks.
    """
    num_block_sums = np.asarray(num_block_sums, dtype=float)
    den_block_sums = np.asarray(den_block_sums, dtype=float)

    # Keep negative denominators, drop only non-finite and exactly-zero denominators
    mask = (den_block_sums != 0) & np.isfinite(num_block_sums) & np.isfinite(den_block_sums)
    if not np.any(mask):
        return BlockJackknifeResult(float("nan"), float("nan"), float("nan"), float("nan"), 0, 0)

    nb = int(mask.sum())
    num_b = num_block_sums[mask]
    den_b = den_block_sums[mask]
    num_total = float(np.sum(num_b))
    den_total = float(np.sum(den_b))
    if den_total == 0:
        return BlockJackknifeResult(float("nan"), float("nan"), float("nan"), float("nan"), nb, int(np.sum(den_b)))

    est = num_total / den_total
    # delete-one estimates
    with np.errstate(divide="ignore", invalid="ignore"):
        est_loo = (num_total - num_b) / (den_total - den_b)
    # Remove degenerate cases where denominator becomes zero after deletion
    valid_loo = np.isfinite(est_loo)
    est_loo = est_loo[valid_loo]
    nb2 = est_loo.size
    if nb2 == 0:
        return BlockJackknifeResult(est, 0.0, float("nan"), float("nan"), nb, int(np.sum(den_b)))
    # standard jackknife variance
    se = float(np.sqrt((nb2 - 1) / nb2 * np.sum((est_loo - est) ** 2)))
    z = float(est / se) if se > 0 else float("nan")
    p = float(math.erfc(abs(z) / math.sqrt(2))) if np.isfinite(z) else float("nan")
    return BlockJackknifeResult(est, se, z, p, nb, int(np.sum(den_b)))


def _aggregate_to_pop_allele_freq(
    calldata_gt: np.ndarray,
    sample_labels: Sequence[str],
    *,
    ancestry: Optional[Union[str, int]] = None,
    snpobj: Optional[Any] = None,
    laiobj: Optional[Any] = None,
) -> Tuple[np.ndarray, np.ndarray, List[str]]:
    """
    Convert sample-level genotypes into per-population allele frequency and count matrices.

    Genotype encoding supported:
        - 3D (n_snps, n_samples, 2): haplotype calls in {0,1}, missing as negative or NaN
        - 2D (n_snps, n_samples): diploid dosages in {0,1,2} or haploid in {0,1}; missing as negative or NaN

    Returns:
        afs: float array (n_snps, n_pops)
        counts: int array (n_snps, n_pops) of called haplotypes per SNP and population
        pops: list of population names in column order
    """
    gt = np.asarray(calldata_gt)
    if gt.ndim not in (2, 3):
        raise ValueError("'calldata_gt' must be 2D or 3D array")

    sample_labels = np.asarray(sample_labels)
    pops, pop_indices = np.unique(sample_labels, return_inverse=True)
    n_snps = gt.shape[0]
    n_pops = pops.size

    # Optional ancestry mask (requires haplotype-level genotypes)
    if ancestry is not None:
        laj = None
        if snpobj is not None and getattr(snpobj, "calldata_lai", None) is not None:
            laj = snpobj.calldata_lai
        if laiobj is not None and laj is None:
            try:
                # Prefer 3D LAI to mask haplotypes
                laj = laiobj.convert_to_snp_level(snpobject=snpobj, lai_format="3D").calldata_lai
            except Exception:
                pass
        if laj is None:
            raise ValueError(
                "Ancestry-specific masking requires SNP-level LAI (provide a LocalAncestryObject via 'laiobj' or ensure 'snpobj.calldata_lai' is set)."
            )
        if gt.ndim != 3:
            raise ValueError("Ancestry-specific masking requires 3D genotype array (n_snps, n_samples, 2).")
        # Normalize LAI to 3D
        laj = np.asarray(laj)
        if laj.ndim == 2:
            n_samples = gt.shape[1]
            try:
                laj = laj.reshape(n_snps, n_samples, 2)
            except Exception:
                raise ValueError("LAI shape is incompatible with genotypes. Expected (n_snps, n_samples*2) to reshape.")
        if laj.ndim != 3:
            raise ValueError("LAI must be 3D (n_snps, n_samples, 2) or 2D (n_snps, n_samples*2).")
        # Apply mask: keep entries matching ancestry; others set to NaN
        mask = (laj.astype(str) == str(ancestry))
        gt = gt.astype(float)
        gt[~mask] = np.nan

    # Compute alt allele counts and haplotype counts per SNP and sample
    if gt.ndim == 3:
        # (n_snps, n_samples, 2)
        g = gt.astype(float)
        # treat values < 0 as missing
        missing = g < 0
        g[missing] = np.nan
        alt_counts_per_sample = np.nansum(g, axis=2)  # sum haplotypes per sample
        hap_count_per_sample = 2 - np.sum(np.isnan(g), axis=2)
    else:
        # (n_snps, n_samples)
        g = gt.astype(float)
        missing = g < 0
        g[missing] = np.nan
        all_nan = np.all(np.isnan(g))
        max_val = np.nan if all_nan else np.nanmax(g)
        if all_nan:
            hap_count_per_sample = np.zeros_like(g)
            alt_counts_per_sample = np.zeros_like(g)
        elif max_val <= 1:
            hap_count_per_sample = np.where(np.isnan(g), 0.0, 1.0)
            alt_counts_per_sample = np.where(np.isnan(g), 0.0, g)
        else:
            hap_count_per_sample = np.where(np.isnan(g), 0.0, 2.0)
            alt_counts_per_sample = np.where(np.isnan(g), 0.0, g)

    # Aggregate per population
    afs = np.zeros((n_snps, n_pops), dtype=float)
    counts = np.zeros((n_snps, n_pops), dtype=float)
    for k in range(n_pops):
        cols = np.where(pop_indices == k)[0]
        if cols.size == 0:
            continue
        counts[:, k] = np.sum(hap_count_per_sample[:, cols], axis=1)
        alt_sum = np.sum(alt_counts_per_sample[:, cols], axis=1)
        with np.errstate(divide="ignore", invalid="ignore"):
            afs[:, k] = np.where(counts[:, k] > 0, alt_sum / counts[:, k], np.nan)
    return afs, counts.astype(int), pops.tolist()


def _prepare_inputs(
    data: Union[Any, Tuple[np.ndarray, np.ndarray, List[str]]],
    sample_labels: Optional[Sequence[str]] = None,
    *,
    ancestry: Optional[Union[str, int]] = None,
    laiobj: Optional[Any] = None,
) -> Tuple[np.ndarray, np.ndarray, List[str]]:
    """
    Normalize inputs to allele frequencies, counts, and optional variant metadata.
    """
    try:
        from snputils.snp.genobj.snpobj import SNPObject  # type: ignore
        is_snpobj = isinstance(data, SNPObject)
    except Exception:
        is_snpobj = False

    if is_snpobj:
        snpobj = data
        if sample_labels is None:
            if snpobj.samples is None:
                raise ValueError("sample_labels must be provided when SNPObject.samples is None")
            sample_labels = snpobj.samples
        afs, counts, pops = _aggregate_to_pop_allele_freq(
            snpobj.calldata_gt,
            sample_labels,
            ancestry=ancestry,
            snpobj=snpobj,
            laiobj=laiobj,
        )
        return afs, counts, pops

    if isinstance(data, tuple) and len(data) == 3:
        afs, counts, pops = data
        return np.asarray(afs), np.asarray(counts), list(pops)

    raise ValueError(
        "data must be either a SNPObject or a tuple (afs, counts, pops) where afs/counts have shape (n_snps, n_pops)"
    )


def _build_blocks(
    n_snps: int,
    blocks: Optional[np.ndarray],
    block_size: int,
) -> Tuple[np.ndarray, np.ndarray]:
    if blocks is not None:
        blocks = np.asarray(blocks)
        if blocks.shape[0] != n_snps:
            raise ValueError("'blocks' must have length n_snps")
        return _compute_block_indices(n_snps, block_labels=blocks)
    return _compute_block_indices(n_snps=n_snps, size=block_size)


def f2(
    data: Union[Any, Tuple[np.ndarray, np.ndarray, List[str]]],
    pop1: Optional[Sequence[str]] = None,
    pop2: Optional[Sequence[str]] = None,
    sample_labels: Optional[Sequence[str]] = None,
    apply_correction: bool = True,
    block_size: int = 5000,
    blocks: Optional[np.ndarray] = None,
    ancestry: Optional[Union[str, int]] = None,
    laiobj: Optional[Any] = None,
    include_self: bool = False,
) -> pd.DataFrame:
    """
    Compute f2-statistics with block-jackknife standard errors.

    Args:
        data: Either a SNPObject or a tuple (afs, counts, pops), where `afs` and `counts` are arrays of
              shape (n_snps, n_pops). If a SNPObject is provided, `sample_labels` are used to aggregate samples to populations.
        pop1, pop2: Populations to compute f2 for. 
            If None:
                - with include_self=False (default), compute only off-diagonal pairs i<j.
                - with include_self=True, compute all pairs including diagonals (i<=j).
        sample_labels: Population label per sample (aligned with SNPObject.samples) when `data` is a SNPObject.
        apply_correction: Apply small-sample correction p*(1-p)/(n-1) per population.
            When True, SNPs with n<=1 in either population are excluded at that SNP.
        block_size: Number of SNPs per jackknife block (default 5000 SNPs). Ignored if `blocks` is provided.
        blocks: Optional explicit block id per SNP. If provided, overrides `block_size`.
        ancestry: Optional ancestry code to mask genotypes to a specific ancestry before aggregation. Requires LAI.
        laiobj: Optional `LocalAncestryObject` used to derive SNP-level LAI if it is not already present on the SNPObject.

    Returns:
        Pandas DataFrame with columns: pop1, pop2, est, se, z, p, n_blocks, n_snps
    """
    afs, counts, pops = _prepare_inputs(data, sample_labels, ancestry=ancestry, laiobj=laiobj)
    n_snps, n_pops = afs.shape
    block_ids, block_lengths = _build_blocks(n_snps, blocks, block_size)
    n_blocks = block_lengths.size

    if pop1 is None and pop2 is None:
        if include_self:
            pop_pairs = [(pops[i], pops[j]) for i in range(n_pops) for j in range(i, n_pops)]
        else:
            pop_pairs = [(pops[i], pops[j]) for i in range(n_pops) for j in range(i + 1, n_pops)]
    else:
        if pop1 is None or pop2 is None:
            raise ValueError("Both pop1 and pop2 must be provided when one is provided")
        if len(pop1) != len(pop2):
            raise ValueError("pop1 and pop2 must have the same length")
        pop_pairs = list(zip(pop1, pop2))

    name_to_idx = {p: i for i, p in enumerate(pops)}

    rows: List[Dict[str, Union[str, float, int]]] = []
    # Precompute per-block index lists for efficiency
    block_bins = [np.where(block_ids == b)[0] for b in range(n_blocks)]

    for p1, p2 in pop_pairs:
        i = name_to_idx[p1]
        j = name_to_idx[p2]
        p_i = afs[:, i]
        p_j = afs[:, j]
        n_i = counts[:, i].astype(float)
        n_j = counts[:, j].astype(float)

        # per-SNP f2 with optional small-sample correction.
        # If apply_correction is True, SNPs with n<=1 for either population are excluded.
        with np.errstate(divide="ignore", invalid="ignore"):
            num = (p_i - p_j) ** 2
            if apply_correction:
                corr_i = np.where(n_i > 1, (p_i * (1.0 - p_i)) / (n_i - 1.0), np.nan)
                corr_j = np.where(n_j > 1, (p_j * (1.0 - p_j)) / (n_j - 1.0), np.nan)
                num = num - corr_i - corr_j
        # SNPs contribute only if the correction was defined (finite) and AFs present
        if apply_correction:
            snp_mask = np.isfinite(num)
        else:
            snp_mask = np.isfinite(num) & (n_i > 0) & (n_j > 0)

        # Aggregate per block: sums of num and counts. Use NaN to mark empty blocks.
        num_block_sums = np.full(n_blocks, np.nan, dtype=float)
        den_block_sums = np.full(n_blocks, np.nan, dtype=float)
        for b, idx in enumerate(block_bins):
            if idx.size == 0:
                continue
            idx2 = idx[snp_mask[idx]]
            if idx2.size == 0:
                continue
            num_block_sums[b] = float(np.nansum(num[idx2]))
            den_block_sums[b] = float(idx2.size)

        res = _jackknife_ratio_from_block_sums(num_block_sums, den_block_sums)
        rows.append(
            {
                "pop1": p1,
                "pop2": p2,
                "est": res.est,
                "se": res.se,
                "z": res.z,
                "p": res.p,
                "n_blocks": res.n_blocks,
                "n_snps": res.n_snps,
            }
        )

    return pd.DataFrame(rows)


def f3(
    data: Union[Any, Tuple[np.ndarray, np.ndarray, List[str]]],
    target: Optional[Sequence[str]] = None,
    ref1: Optional[Sequence[str]] = None,
    ref2: Optional[Sequence[str]] = None,
    sample_labels: Optional[Sequence[str]] = None,
    apply_correction: bool = True,
    block_size: int = 5000,
    blocks: Optional[np.ndarray] = None,
    ancestry: Optional[Union[str, int]] = None,
    laiobj: Optional[Any] = None,
) -> pd.DataFrame:
    """
    Compute f3-statistics f3(target; ref1, ref2) with block-jackknife SE.

    - `block_size` is the number of SNPs per jackknife block (default 5000 SNPs). Ignored if `blocks` is provided.
    - If `target`, `ref1`, and `ref2` are all None, compute f3 for all combinations where each role can be any population.
    - If `ancestry` is provided, genotypes will be masked to the specified ancestry using LAI before aggregation.
    - If `apply_correction` is True, subtract the finite sample term p_t*(1-p_t)/(n_t-1) from the per-SNP product.
        When True, SNPs with n_t<=1 are excluded.
    """
    afs, counts, pops = _prepare_inputs(data, sample_labels, ancestry=ancestry, laiobj=laiobj)
    n_snps, _ = afs.shape
    block_ids, block_lengths = _build_blocks(n_snps, blocks, block_size)
    n_blocks = block_lengths.size

    if target is None and ref1 is None and ref2 is None:
        triples = [(a, b, c) for a in pops for b in pops for c in pops]
    else:
        if target is None or ref1 is None or ref2 is None:
            raise ValueError("target, ref1, and ref2 must all be provided if any is provided")
        if not (len(target) == len(ref1) == len(ref2)):
            raise ValueError("target, ref1, ref2 must have the same length")
        triples = list(zip(target, ref1, ref2))

    name_to_idx = {p: i for i, p in enumerate(pops)}
    block_bins = [np.where(block_ids == b)[0] for b in range(n_blocks)]
    rows: List[Dict[str, Union[str, float, int]]] = []

    for t, r1, r2 in triples:
        it = name_to_idx[t]
        i1 = name_to_idx[r1]
        i2 = name_to_idx[r2]
        pt = afs[:, it]
        p1 = afs[:, i1]
        p2 = afs[:, i2]
        nt = counts[:, it].astype(float)
        n1 = counts[:, i1].astype(float)
        n2 = counts[:, i2].astype(float)

        with np.errstate(invalid="ignore", divide="ignore"):
            num = (pt - p1) * (pt - p2)
            if apply_correction:
                corr_t = np.where(nt > 1, (pt * (1.0 - pt)) / (nt - 1.0), np.nan)
                num = num - corr_t
        if apply_correction:
            # Require a valid correction on the target, references can be n>0
            snp_mask = np.isfinite(num) & (n1 > 0) & (n2 > 0)
        else:
            snp_mask = np.isfinite(num) & (nt > 0) & (n1 > 0) & (n2 > 0)

        num_block_sums = np.full(n_blocks, np.nan, dtype=float)
        den_block_sums = np.full(n_blocks, np.nan, dtype=float)
        for b, idx in enumerate(block_bins):
            if idx.size == 0:
                continue
            idx2 = idx[snp_mask[idx]]
            if idx2.size == 0:
                continue
            num_block_sums[b] = float(np.nansum(num[idx2]))
            den_block_sums[b] = float(idx2.size)

        res = _jackknife_ratio_from_block_sums(num_block_sums, den_block_sums)
        rows.append(
            {
                "target": t,
                "ref1": r1,
                "ref2": r2,
                "est": res.est,
                "se": res.se,
                "z": res.z,
                "p": res.p,
                "n_blocks": res.n_blocks,
                "n_snps": res.n_snps,
            }
        )

    return pd.DataFrame(rows)


def f4(
    data: Union[Any, Tuple[np.ndarray, np.ndarray, List[str]]],
    a: Optional[Sequence[str]] = None,
    b: Optional[Sequence[str]] = None,
    c: Optional[Sequence[str]] = None,
    d: Optional[Sequence[str]] = None,
    sample_labels: Optional[Sequence[str]] = None,
    block_size: int = 5000,
    blocks: Optional[np.ndarray] = None,
    ancestry: Optional[Union[str, int]] = None,
    laiobj: Optional[Any] = None,
) -> pd.DataFrame:
    """
    Compute f4-statistics f4(a, b; c, d) with block-jackknife SE.

    - `block_size` is the number of SNPs per jackknife block (default 5000 SNPs). Ignored if `blocks` is provided.
    - If `ancestry` is provided, genotypes will be masked to the specified ancestry using LAI before aggregation.
    """
    afs, counts, pops = _prepare_inputs(data, sample_labels, ancestry=ancestry, laiobj=laiobj)
    n_snps, _ = afs.shape
    block_ids, block_lengths = _build_blocks(n_snps, blocks, block_size)
    n_blocks = block_lengths.size

    if a is None and b is None and c is None and d is None:
        quads = [(w, x, y, z) for w in pops for x in pops for y in pops for z in pops]
    else:
        if a is None or b is None or c is None or d is None:
            raise ValueError("a, b, c, d must all be provided if any is provided")
        if not (len(a) == len(b) == len(c) == len(d)):
            raise ValueError("a, b, c, d must have the same length")
        quads = list(zip(a, b, c, d))

    name_to_idx = {p: i for i, p in enumerate(pops)}
    block_bins = [np.where(block_ids == b)[0] for b in range(n_blocks)]
    rows: List[Dict[str, Union[str, float, int]]] = []

    for pa, pb, pc, dpop in quads:
        ia = name_to_idx[pa]
        ib = name_to_idx[pb]
        ic = name_to_idx[pc]
        id_ = name_to_idx[dpop]

        A = afs[:, ia]
        B = afs[:, ib]
        C = afs[:, ic]
        D = afs[:, id_]
        na = counts[:, ia].astype(float)
        nb = counts[:, ib].astype(float)
        nc = counts[:, ic].astype(float)
        nd = counts[:, id_].astype(float)

        with np.errstate(invalid="ignore"):
            num = (A - B) * (C - D)
        snp_mask = np.isfinite(num) & (na > 0) & (nb > 0) & (nc > 0) & (nd > 0)

        num_block_sums = np.full(n_blocks, np.nan, dtype=float)
        den_block_sums = np.full(n_blocks, np.nan, dtype=float)
        for b_idx, idx in enumerate(block_bins):
            if idx.size == 0:
                continue
            idx2 = idx[snp_mask[idx]]
            if idx2.size == 0:
                continue
            num_block_sums[b_idx] = float(np.nansum(num[idx2]))
            den_block_sums[b_idx] = float(idx2.size)

        res = _jackknife_ratio_from_block_sums(num_block_sums, den_block_sums)
        rows.append(
            {
                "a": pa,
                "b": pb,
                "c": pc,
                "d": dpop,
                "est": res.est,
                "se": res.se,
                "z": res.z,
                "p": res.p,
                "n_blocks": res.n_blocks,
                "n_snps": res.n_snps,
            }
        )

    return pd.DataFrame(rows)


def d_stat(
    data: Union[Any, Tuple[np.ndarray, np.ndarray, List[str]]],
    a: Optional[Sequence[str]] = None,
    b: Optional[Sequence[str]] = None,
    c: Optional[Sequence[str]] = None,
    d: Optional[Sequence[str]] = None,
    sample_labels: Optional[Sequence[str]] = None,
    block_size: int = 5000,
    blocks: Optional[np.ndarray] = None,
    ancestry: Optional[Union[str, int]] = None,
    laiobj: Optional[Any] = None,
) -> pd.DataFrame:
    """
    Compute D-statistics D(a, b; c, d) as ratio of sums:
        D = sum_l (A-B)(C-D)  /  sum_l (A+B-2AB)(C+D-2CD)
    with delete-one block jackknife SE.

    - `block_size` is the number of SNPs per jackknife block (default 5000 SNPs). Ignored if `blocks` is provided.
    - If `ancestry` is provided, genotypes will be masked to the specified ancestry using LAI before aggregation.
    """
    afs, counts, pops = _prepare_inputs(data, sample_labels, ancestry=ancestry, laiobj=laiobj)
    n_snps, _ = afs.shape
    block_ids, block_lengths = _build_blocks(n_snps, blocks, block_size)
    n_blocks = block_lengths.size

    if a is None and b is None and c is None and d is None:
        quads = [(w, x, y, z) for w in pops for x in pops for y in pops for z in pops]
    else:
        if a is None or b is None or c is None or d is None:
            raise ValueError("a, b, c, d must all be provided if any is provided")
        if not (len(a) == len(b) == len(c) == len(d)):
            raise ValueError("a, b, c, d must have the same length")
        quads = list(zip(a, b, c, d))

    name_to_idx = {p: i for i, p in enumerate(pops)}
    block_bins = [np.where(block_ids == b)[0] for b in range(n_blocks)]
    rows: List[Dict[str, Union[str, float, int]]] = []

    for pa, pb, pc, dpop in quads:
        ia = name_to_idx[pa]
        ib = name_to_idx[pb]
        ic = name_to_idx[pc]
        id_ = name_to_idx[dpop]

        A = afs[:, ia]
        B = afs[:, ib]
        C = afs[:, ic]
        D = afs[:, id_]
        na = counts[:, ia].astype(float)
        nb = counts[:, ib].astype(float)
        nc = counts[:, ic].astype(float)
        nd = counts[:, id_].astype(float)

        with np.errstate(invalid="ignore"):
            num = (A - B) * (C - D)
            den = (A + B - 2 * A * B) * (C + D - 2 * C * D)

        snp_mask = np.isfinite(num) & np.isfinite(den) & (na > 0) & (nb > 0) & (nc > 0) & (nd > 0)

        num_block_sums = np.full(n_blocks, np.nan, dtype=float)
        den_block_sums = np.full(n_blocks, np.nan, dtype=float)
        ct_block_sums = np.zeros(n_blocks, dtype=int)
        for b_idx, idx in enumerate(block_bins):
            if idx.size == 0:
                continue
            idx2 = idx[snp_mask[idx]]
            if idx2.size == 0:
                continue
            num_block_sums[b_idx] = float(np.nansum(num[idx2]))
            # drop near-zero denominators within a block for stability
            block_den = float(np.nansum(den[idx2]))
            den_block_sums[b_idx] = block_den if abs(block_den) > 1e-12 else np.nan
            ct_block_sums[b_idx] = int(idx2.size)
            
        res = _jackknife_ratio_from_block_sums(num_block_sums, den_block_sums)
        rows.append(
            {
                "a": pa,
                "b": pb,
                "c": pc,
                "d": dpop,
                "est": res.est,
                "se": res.se,
                "z": res.z,
                "p": res.p,
                "n_blocks": res.n_blocks,
                "n_snps": int(ct_block_sums.sum()),
            }
        )

    return pd.DataFrame(rows)


def f4_ratio(
    data: Union[Any, Tuple[np.ndarray, np.ndarray, List[str]]],
    num: Sequence[Tuple[str, str, str, str]],
    den: Sequence[Tuple[str, str, str, str]],
    sample_labels: Optional[Sequence[str]] = None,
    block_size: int = 5000,
    blocks: Optional[np.ndarray] = None,
    ancestry: Optional[Union[str, int]] = None,
    laiobj: Optional[Any] = None,
) -> pd.DataFrame:
    """
    Compute f4-ratio statistics as ratio of two f4-statistics with block-jackknife SE.

    Args:
        num: Sequence of quadruples (a, b, c, d) for numerator f4(a, b; c, d)
        den: Sequence of quadruples (a, b, c, d) for denominator f4(a, b; c, d)

    Notes:
        - `block_size` is the number of SNPs per jackknife block (default 5000 SNPs). Ignored if `blocks` is provided.
        - If `ancestry` is provided, genotypes will be masked to the specified ancestry using LAI before aggregation.
    """
    if len(num) != len(den):
        raise ValueError("'num' and 'den' must have the same length")

    afs, counts, pops = _prepare_inputs(data, sample_labels, ancestry=ancestry, laiobj=laiobj)
    n_snps, _ = afs.shape
    block_ids, block_lengths = _build_blocks(n_snps, blocks, block_size)
    n_blocks = block_lengths.size
    block_bins = [np.where(block_ids == b)[0] for b in range(n_blocks)]
    name_to_idx = {p: i for i, p in enumerate(pops)}

    rows: List[Dict[str, Union[str, float, int]]] = []
    for (na, nb, nc, nd), (da, db, dc, dd) in zip(num, den):
        ia, ib, ic, id_ = name_to_idx[na], name_to_idx[nb], name_to_idx[nc], name_to_idx[nd]
        ja, jb, jc, jd = name_to_idx[da], name_to_idx[db], name_to_idx[dc], name_to_idx[dd]

        A, B, C, D = afs[:, ia], afs[:, ib], afs[:, ic], afs[:, id_]
        E, F, G, H = afs[:, ja], afs[:, jb], afs[:, jc], afs[:, jd]
        nA, nB, nC, nD = counts[:, ia], counts[:, ib], counts[:, ic], counts[:, id_]
        nE, nF, nG, nH = counts[:, ja], counts[:, jb], counts[:, jc], counts[:, jd]

        with np.errstate(invalid="ignore"):
            num_snp = (A - B) * (C - D)
            den_snp = (E - F) * (G - H)
        mask_num = np.isfinite(num_snp) & (nA > 0) & (nB > 0) & (nC > 0) & (nD > 0)
        mask_den = np.isfinite(den_snp) & (nE > 0) & (nF > 0) & (nG > 0) & (nH > 0)
        mask_both = mask_num & mask_den

        num_block_sums = np.full(n_blocks, np.nan, dtype=float)
        den_block_sums = np.full(n_blocks, np.nan, dtype=float)
        ct_block_sums = np.zeros(n_blocks, dtype=int)
        for b_idx, idx in enumerate(block_bins):
            if idx.size == 0:
                continue
            idx2 = idx[mask_both[idx]]
            if idx2.size == 0:
                continue
            num_block_sums[b_idx] = float(np.nansum(num_snp[idx2]))
            block_den = float(np.nansum(den_snp[idx2]))
            den_block_sums[b_idx] = block_den if abs(block_den) > 1e-12 else np.nan
            ct_block_sums[b_idx] = int(idx2.size)
            
        res = _jackknife_ratio_from_block_sums(num_block_sums, den_block_sums)
        rows.append(
            {
                "num": f"({na},{nb};{nc},{nd})",
                "den": f"({da},{db};{dc},{dd})",
                "est": res.est,
                "se": res.se,
                "z": res.z,
                "p": res.p,
                "n_blocks": res.n_blocks,
                "n_snps": int(ct_block_sums.sum()),
            }
        )

    return pd.DataFrame(rows)


def fst(
    data: Union[Any, Tuple[np.ndarray, np.ndarray, List[str]]],
    pop1: Optional[Sequence[str]] = None,
    pop2: Optional[Sequence[str]] = None,
    *,
    method: str = "hudson",
    sample_labels: Optional[Sequence[str]] = None,
    block_size: int = 5000,
    blocks: Optional[np.ndarray] = None,
    ancestry: Optional[Union[str, int]] = None,
    laiobj: Optional[Any] = None,
    include_self: bool = False,
) -> pd.DataFrame:
    """
    Pairwise F_ST with delete-one block jackknife SE.

    Methods:
      - 'hudson' (a.k.a. "ratio of averages" following Hudson 1992 / Bhatia 2013):
            per-SNP num = d_xy - 0.5*(pi_x + pi_y)
            per-SNP den = d_xy
        where d_xy = p_x*(1-p_y) + p_y*(1-p_x) and
              pi_x = 2*p_x*(1-p_x) * n_x/(n_x - 1) (unbiased within-pop diversity on haplotypes).
      - 'weir_cockerham' (Weir & Cockerham 1984; θ):
            compute per-SNP variance components a,b,c (for r=2) from allele freqs and haplotype counts:
                n1, n2 = haplotype counts; p1, p2 = allele freqs
                n  = n1 + n2
                n_bar = n / 2
                p_bar = (n1*p1 + n2*p2) / n
                s2    = (n1*(p1 - p_bar)**2 + n2*(p2 - p_bar)**2) / n_bar
                h1, h2 = 2*p1*(1-p1), 2*p2*(1-p2)
                h_bar = 0.5*(h1 + h2)
                n_c   = (n - (n1**2 + n2**2)/n)  # equals 2*n1*n2/n for r=2
                a = (n_bar / n_c) * (s2 - (p_bar*(1 - p_bar) - 0.5*s2 - 0.25*h_bar) / (n_bar - 1))
                b = (n_bar / (n_bar - 1)) * (p_bar*(1 - p_bar) - 0.5*s2 - ((2*n_bar - 1)/(4*n_bar))*h_bar)
                c = 0.5 * h_bar
            and use num=a, den=(a+b+c), then ratio-of-sums jackknife.

    Notes:
      * Inputs are the same as f2/f3/f4: either SNPObject or (afs, counts, pops).
      * For WC we use expected heterozygosity h_i = 2 p_i (1 - p_i) from allele freqs.
      * SNPs with n<=1 in either pop or with invalid denominators are ignored.
    """
    method = str(method).strip().lower().replace(" ", "_").replace("-", "_")
    if method in ("wc", "weir", "weir_cockerham", "weir-cockerham"):
        method = "weir_cockerham"
    elif method in ("h", "hudson", "bhatia", "ratio", "ratio_of_averages", "ratio-of-averages"):
        method = "hudson"
    elif method not in ("hudson", "weir_cockerham"):
        # only raise if it's neither a known alias nor a canonical name
        raise ValueError(f"Unknown method for fst: {method!r}")

    afs, counts, pops = _prepare_inputs(data, sample_labels, ancestry=ancestry, laiobj=laiobj)
    n_snps, n_pops = afs.shape
    block_ids, block_lengths = _build_blocks(n_snps, blocks, block_size)
    n_blocks = block_lengths.size

    if pop1 is None and pop2 is None:
        if include_self:
            pairs = [(pops[i], pops[j]) for i in range(n_pops) for j in range(i, n_pops)]
        else:
            pairs = [(pops[i], pops[j]) for i in range(n_pops) for j in range(i + 1, n_pops)]
    else:
        if pop1 is None or pop2 is None or len(pop1) != len(pop2):
            raise ValueError("pop1 and pop2 must both be provided and of equal length")
        pairs = list(zip(pop1, pop2))

    name_to_idx = {p: i for i, p in enumerate(pops)}
    block_bins = [np.where(block_ids == b)[0] for b in range(n_blocks)]

    out_rows: List[Dict[str, Union[str, float, int]]] = []

    for pA, pB in pairs:
        i = name_to_idx[pA]
        j = name_to_idx[pB]
        p1 = afs[:, i].astype(float)
        p2 = afs[:, j].astype(float)
        n1 = counts[:, i].astype(float)
        n2 = counts[:, j].astype(float)

        valid = np.isfinite(p1) & np.isfinite(p2)

        if method == "hudson":
            # d_xy and within-pop diversities (unbiased, haplotype-based)
            d_xy = p1 * (1.0 - p2) + p2 * (1.0 - p1)
            with np.errstate(divide="ignore", invalid="ignore"):
                pi1 = 2.0 * p1 * (1.0 - p1) * (n1 / (n1 - 1.0))
                pi2 = 2.0 * p2 * (1.0 - p2) * (n2 / (n2 - 1.0))
            num_snp = d_xy - 0.5 * (pi1 + pi2)
            den_snp = d_xy
            snp_mask = valid & (n1 > 1) & (n2 > 1) & np.isfinite(num_snp) & np.isfinite(den_snp)
        else:
            # Weir-Cockerham θ components (r=2)
            n = n1 + n2
            with np.errstate(divide="ignore", invalid="ignore"):
                n_bar = n / 2.0
                p_bar = np.where(n > 0, (n1 * p1 + n2 * p2) / n, np.nan)
                s2 = (n1 * (p1 - p_bar) ** 2 + n2 * (p2 - p_bar) ** 2) / n_bar
                h1 = 2.0 * p1 * (1.0 - p1)
                h2 = 2.0 * p2 * (1.0 - p2)
                h_bar = 0.5 * (h1 + h2)
                n_c = n - (n1 * n1 + n2 * n2) / np.where(n > 0, n, np.nan)  # == 2*n1*n2/n
                # components
                a = (n_bar / n_c) * (s2 - (p_bar * (1.0 - p_bar) - 0.5 * s2 - 0.25 * h_bar) / (n_bar - 1.0))
                b = (n_bar / (n_bar - 1.0)) * (p_bar * (1.0 - p_bar) - 0.5 * s2 - ((2.0 * n_bar - 1.0) / (4.0 * n_bar)) * h_bar)
                c = 0.5 * h_bar
                num_snp = a
                den_snp = a + b + c
            # Need at least 2 haplotypes per pop and well-defined denominators
            snp_mask = valid & (n1 > 1) & (n2 > 1) & np.isfinite(num_snp) & np.isfinite(den_snp)

        # Aggregate by blocks
        num_block_sums = np.full(n_blocks, np.nan, dtype=float)
        den_block_sums = np.full(n_blocks, np.nan, dtype=float)
        ct_block_sums = np.zeros(n_blocks, dtype=int)

        for b_idx, idx in enumerate(block_bins):
            if idx.size == 0:
                continue
            idx2 = idx[snp_mask[idx]]
            if idx2.size == 0:
                continue
            ns = float(np.nansum(num_snp[idx2]))
            ds = float(np.nansum(den_snp[idx2]))
            # drop numerically-zero denominators for stability
            den_block_sums[b_idx] = ds if abs(ds) > 1e-12 else np.nan
            num_block_sums[b_idx] = ns
            ct_block_sums[b_idx] = int(idx2.size)

        res = _jackknife_ratio_from_block_sums(num_block_sums, den_block_sums)
        out_rows.append(
            {
                "pop1": pA,
                "pop2": pB,
                "method": method,
                "est": res.est,
                "se": res.se,
                "z": res.z,
                "p": res.p,
                "n_blocks": res.n_blocks,
                "n_snps": int(ct_block_sums.sum()),
            }
        )

    return pd.DataFrame(out_rows)



__all__ = [
    "f2",
    "f3",
    "f4",
    "d_stat",
    "f4_ratio",
    "fst",
]





import numpy as np

from snputils import VCFReader, BEDReader, PGENReader


# TODO: Fails with KeyError: 'calldata/GT' (genotypes = vcf_dict["calldata/GT"].astype(np.int8))
# def test_vcf_only_samples(data_path, snpobj_vcf):
#     snpobj_vcf_only_samples = VCFReader(data_path + "/subset.vcf").read(fields=["IID"])
#     assert np.array_equal(snpobj_vcf_only_samples.samples, snpobj_vcf.samples)


def test_bed_only_samples(data_path, snpobj_bed):
    snpobj_bed_only_samples = BEDReader(data_path + "/bed/subset").read(fields=["IID"])
    assert np.array_equal(snpobj_bed_only_samples.samples, snpobj_bed.samples)


def test_pgen_only_samples(data_path, snpobj_pgen):
    snpobj_pgen_only_samples = PGENReader(data_path + "/pgen/subset").read(fields=["IID"])
    assert np.array_equal(snpobj_pgen_only_samples.samples, snpobj_pgen.samples)


# TODO: Fails with KeyError: 'calldata/GT' (genotypes = vcf_dict["calldata/GT"].astype(np.int8))
# def test_vcf_only_variants(data_path, snpobj_vcf):
#     snpobj_vcf_only_variants = VCFReader(data_path + "/subset.vcf").read(fields=["ID"])
#     assert np.array_equal(snpobj_vcf_only_variants.variants_id, snpobj_vcf.variants_id)


def test_bed_only_variants(data_path, snpobj_bed):
    snpobj_bed_only_variants = BEDReader(data_path + "/bed/subset").read(fields=["ID"])
    assert np.array_equal(snpobj_bed_only_variants.variants_id, snpobj_bed.variants_id)


def test_pgen_only_variants(data_path, snpobj_pgen):
    snpobj_pgen_only_variants = PGENReader(data_path + "/pgen/subset").read(fields=["ID"])
    assert np.array_equal(snpobj_pgen_only_variants.variants_id, snpobj_pgen.variants_id)


# TODO: Fails with KeyError: 'samples' (samples=vcf_dict["samples"],)
# def test_vcf_only_gt(data_path, snpobj_vcf):
#     snpobj_vcf_only_gt = VCFReader(data_path + "/subset.vcf").read(fields=["GT"])
#     assert np.array_equal(snpobj_vcf_only_gt.calldata_gt, snpobj_vcf.calldata_gt)


def test_bed_only_gt(data_path, snpobj_bed):
    snpobj_bed_only_gt = BEDReader(data_path + "/bed/subset").read(fields=["GT"])
    assert np.array_equal(snpobj_bed_only_gt.calldata_gt, snpobj_bed.calldata_gt)


def test_pgen_only_gt(data_path, snpobj_pgen):
    snpobj_pgen_only_gt = PGENReader(data_path + "/pgen/subset").read(fields=["GT"])
    assert np.array_equal(snpobj_pgen_only_gt.calldata_gt, snpobj_pgen.calldata_gt)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib

def reorder_admixture(Q_mat):
    """
    Reorder Q_mat rows so that rows are grouped by each sample's dominant ancestry,
    and columns are sorted by descending average ancestry proportion.
    """
    n_samples, K = Q_mat.shape
    
    # Reorder columns by descending average proportion
    col_means = Q_mat.mean(axis=0)
    col_order = np.argsort(col_means)[::-1]   # largest first
    Q_cols_sorted = Q_mat[:, col_order]
    
    # Group samples by whichever column is argmax
    row_groups = []
    boundary_list = [0]
    argmax_all = np.argmax(Q_cols_sorted, axis=1)
    for k in range(K):
        rows_k = np.where(argmax_all == k)[0]
        # Sort these rows by their proportion in col k
        rows_k_sorted = rows_k[np.argsort(Q_cols_sorted[rows_k, k])[::-1]]
        row_groups.append(rows_k_sorted)
        boundary_list.append(boundary_list[-1] + len(rows_k_sorted))
    
    # Combine them into one final row order
    row_order = np.concatenate(row_groups) if row_groups else np.arange(n_samples)
    Q_mat_sorted = Q_cols_sorted[row_order, :]
    
    return Q_mat_sorted, row_order, boundary_list, col_order

def plot_admixture(ax, Q_mat_sorted, boundary_list, col_order=None, colors=None, show_boundaries=True, show_axes_labels=True, show_ticks=True, set_limits=True, minimal=False):
    """
    Plot a structure-style bar chart of Q_mat_sorted in the given Axes ax.
    If colors is not None, it should be a list or array of length K.
    If col_order is not None, colors are reordered according to col_order.

    Optional controls:
    - show_boundaries (bool): draw vertical lines at group boundaries. Default True.
    - show_axes_labels (bool): set X/Y axis labels. Default True.
    - show_ticks (bool): show axis ticks. Default True.
    - set_limits (bool): set xlim and ylim to [0, n_samples-1] and [0,1]. Default True.
    - minimal (bool): if True, overrides to disable boundaries, labels, ticks, limits and hides spines.
    """
    n_samples, K = Q_mat_sorted.shape

    # Minimal overrides
    if minimal:
        show_boundaries = False
        show_axes_labels = False
        show_ticks = False
        set_limits = False

    # If we have a specific color list and a col_order, reorder the colors to match the columns
    if (colors is not None) and (col_order is not None):
        colors = [colors[idx] for idx in col_order]

    # cumulative sum across columns for stacked fill
    Q_cum = np.cumsum(Q_mat_sorted, axis=1)
    # Use step='post' with padded x/y so the last bar renders fully and no thin band appears
    x_edges = np.arange(n_samples + 1)
    Q_pad = np.vstack([Q_cum, Q_cum[-1]])

    # fill-between for a stacked bar effect
    for j in range(K):
        c = colors[j] if (colors is not None) else None
        lower = Q_pad[:, j - 1] if j > 0 else np.zeros(n_samples + 1)
        upper = Q_pad[:, j]
        ax.fill_between(
            x_edges,
            lower,
            upper,
            step="post",
            color=c,
            linewidth=0,
            edgecolor='none',
        )

    # Vertical lines for group boundaries
    if show_boundaries:
        for boundary in boundary_list:
            ax.axvline(boundary, color='black', ls='--', lw=1.0)

    if set_limits:
        ax.set_xlim(0, n_samples)
        ax.set_ylim(0, 1)

    if show_axes_labels:
        ax.set_xlabel("Samples")
        ax.set_ylabel("Ancestry Proportion")

    if not show_ticks:
        ax.set_xticks([])
        ax.set_yticks([])

    if minimal:
        # Hide all spines for a clean, minimal appearance
        for spine in ax.spines.values():
            spine.set_visible(False)



import numpy as np


def create_benchmark_test(benchmark, reader_func, file_path, name, ref_array, memory_profile):
    if memory_profile:
        from memory_profiler import memory_usage

        def run_with_memory_profile():
            def read_file():
                return reader_func(file_path)

            mem_usage, (result,) = memory_usage(
                (read_file, [], {}),
                retval=True,
                interval=0.1,
                include_children=True
            )
            return result, max(mem_usage)

        # Run benchmark with memory profiling
        result, max_mem = benchmark(run_with_memory_profile)
        benchmark.extra_info['max_memory_mb'] = max_mem

    else:
        # Run benchmark without memory profiling
        result = benchmark(reader_func, file_path)
        benchmark.extra_info['max_memory_mb'] = None

    # Verify output (not timed)
    assert np.array_equal(result, ref_array), f"Output does not match reference for {name}"



import pathlib
import numpy as np
import copy
from typing import Optional, Dict, List, Union

from snputils.snp.genobj.snpobj import SNPObject
from snputils.ancestry.genobj.local import LocalAncestryObject
from ._utils.mds_distance import distance_mat, mds_transform
from ._utils.gen_tools import process_calldata_gt, process_labels_weights


class maasMDS:
    """
    A class for performing multiple array ancestry-specific multidimensional scaling (maasMDS) on SNP data.

    The maasMDS class focuses on genotype segments from the ancestry of interest when the `is_masked` flag is set to `True`. It offers 
    flexible processing options, allowing either separate handling of masked haplotype strands or combining (averaging) strands into a 
    single composite representation for each individual. Moreover, the analysis can be performed on individual-level data, group-level SNP 
    frequencies, or a combination of both.

    This class supports both separate and averaged strand processing for SNP data. If the `snpobj`, 
    `laiobj`, `labels_file`, and `ancestry` parameters are all provided during instantiation, 
    the `fit_transform` method will be automatically called, applying the specified maasMDS method to transform 
    the data upon instantiation.
    """
    def __init__(
            self, 
            snpobj: Optional['SNPObject'] = None,
            laiobj: Optional['LocalAncestryObject'] = None,
            labels_file: Optional[str] = None,
            ancestry: Optional[Union[int, str]] = None,
            is_masked: bool = True,
            average_strands: bool = False,
            force_nan_incomplete_strands: bool = False,
            is_weighted: bool = False,
            groups_to_remove: Dict[int, List[str]] = {},
            min_percent_snps: float = 4,
            group_snp_frequencies_only: bool = True,
            save_masks: bool = False,
            load_masks: bool = False,
            masks_file: Union[str, pathlib.Path] = 'masks.npz',
            distance_type: str = 'AP',
            n_components: int = 2,
            rsid_or_chrompos: int = 2
        ):
        """
        Args:
            snpobj (SNPObject, optional): 
                A SNPObject instance.
            laiobj (LAIObject, optional): 
                A LAIObject instance.
            labels_file (str, optional): 
                Path to the labels file in .tsv format. The first column, `indID`, contains the individual identifiers, and the second 
                column, `label`, specifies the groups for all individuals. If `is_weighted=True`, a `weight` column with individual 
                weights is required. Optionally, `combination` and `combination_weight` columns can specify sets of individuals to be 
                combined into groups, with respective weights.
            ancestry (int or str, optional): 
                Ancestry for which dimensionality reduction is to be performed. Ancestry counter starts at `0`. The ancestry input can be:
                - An integer (e.g., 0, 1, 2).
                - A string representation of an integer (e.g., '0', '1').
                - A string matching one of the ancestry map values (e.g., 'Africa').
            is_masked (bool, default=True): 
                If `True`, applies ancestry-specific masking to the genotype matrix, retaining only genotype data 
                corresponding to the specified `ancestry`. If `False`, uses the full, unmasked genotype matrix.
            average_strands (bool, default=False): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
            force_nan_incomplete_strands (bool): 
                If `True`, sets the result to NaN if either haplotype in a pair is NaN. 
                Otherwise, computes the mean while ignoring NaNs (e.g., 0|NaN -> 0, 1|NaN -> 1).
            is_weighted (bool, default=False): 
                True if weights are provided in the labels file, or False otherwise.
            groups_to_remove (dict of int to list of str, default={}): 
                Dictionary specifying groups to exclude from analysis. Keys are array numbers, and values are 
                lists of groups to remove for each array.
                Example: `{1: ['group1', 'group2'], 2: [], 3: ['group3']}`.
            min_percent_snps (float, default=4.0): 
                Minimum percentage of SNPs to be known in an individual for an individual to be included in the analysis. 
                All individuals with fewer percent of unmasked SNPs than this threshold will be excluded.
            group_snp_frequencies_only (bool, default=True):
                If True, maasMDS is performed exclusively on group-level SNP frequencies, ignoring individual-level data. This applies when `is_weighted` is 
                set to True and a `combination` column is provided in the `labels_file`,  meaning individuals are aggregated into groups based on their assigned 
                labels. If False, maasMDS is performed on individual-level SNP data alone or on both individual-level and group-level SNP frequencies when 
                `is_weighted` is True and a `combination` column is provided.
            save_masks (bool, default=False): 
                True if the masked matrices are to be saved in a `.npz` file, or False otherwise.
            load_masks (bool, default=False): 
                True if the masked matrices are to be loaded from a pre-existing `.npz` file specified by `masks_file`, or False otherwise.
            masks_file (str or pathlib.Path, default='masks.npz'): 
                Path to the `.npz` file used for saving/loading masked matrices.
            distance_type (str, default='AP'): 
                Distance metric to use. Options to choose from are: 'Manhattan', 'RMS' (Root Mean Square), 'AP' (Average Pairwise).
                If `average_strands=True`, use 'distance_type=AP'.
            n_components (int, default=2): 
                The number of principal components.
            rsid_or_chrompos (int, default=2): 
                Format indicator for SNP IDs in the SNP data. Use 1 for `rsID` format or 2 for `chromosome_position`.
        """
        self.__snpobj = snpobj
        self.__laiobj = laiobj
        self.__labels_file = labels_file
        self.__ancestry = self._define_ancestry(ancestry, laiobj.ancestry_map) if laiobj is not None else None
        self.__is_masked = is_masked
        self.__average_strands = average_strands
        self.__force_nan_incomplete_strands = force_nan_incomplete_strands
        self.__groups_to_remove = groups_to_remove
        self.__min_percent_snps = min_percent_snps
        self.__group_snp_frequencies_only = group_snp_frequencies_only
        self.__is_weighted = is_weighted
        self.__save_masks = save_masks
        self.__load_masks = load_masks
        self.__masks_file = masks_file
        self.__distance_type = distance_type
        self.__n_components = n_components
        self.__rsid_or_chrompos = rsid_or_chrompos
        self.__X_new_ = None  # Store transformed SNP data
        self.__haplotypes_ = None  # Store haplotypes after filtering if min_percent_snps > 0
        self.__samples_ = None  # Store samples after filtering if min_percent_snps > 0
        self.__variants_id_ = None  # Store variants ID (after filtering SNPs not in laiobj)

        # Fit and transform if a `snpobj`, `laiobj`, `labels_file`, and `ancestry` are provided
        if self.snpobj is not None and self.laiobj is not None and self.labels_file is not None and self.ancestry is not None:
            self.fit_transform(snpobj, laiobj, labels_file, ancestry)

    def __getitem__(self, key):
        """
        To access an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')

    def __setitem__(self, key, value):
        """
        To set an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            setattr(self, key, value)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')
        
    def copy(self) -> 'maasMDS':
        """
        Create and return a copy of `self`.

        Returns:
            **maasMDS:** 
                A new instance of the current object.
        """
        return copy.copy(self)

    @property
    def snpobj(self) -> Optional['SNPObject']:
        """
        Retrieve `snpobj`.
        
        Returns:
            **SNPObject:** A SNPObject instance.
        """
        return self.__snpobj

    @snpobj.setter
    def snpobj(self, x: 'SNPObject') -> None:
        """
        Update `snpobj`.
        """
        self.__snpobj = x

    @property
    def laiobj(self) -> Optional['LocalAncestryObject']:
        """
        Retrieve `laiobj`.
        
        Returns:
            **LocalAncestryObject:** A LAIObject instance.
        """
        return self.__laiobj

    @laiobj.setter
    def laiobj(self, x: 'LocalAncestryObject') -> None:
        """
        Update `laiobj`.
        """
        self.__laiobj = x

    @property
    def labels_file(self) -> Optional[str]:
        """
        Retrieve `labels_file`.
        
        Returns:
            **str:** 
                Path to the labels file in `.tsv` format.
        """
        return self.__labels_file

    @labels_file.setter
    def labels_file(self, x: str) -> None:
        """
        Update `labels_file`.
        """
        self.__labels_file = x

    @property
    def ancestry(self) -> Optional[int]:
        """
        Retrieve `ancestry`.
        
        Returns:
            **int:** Ancestry index for which dimensionality reduction is to be performed. Ancestry counter starts at `0`.
        """
        return self.__ancestry

    @ancestry.setter
    def ancestry(self, x: Union[int, str]) -> None:
        """
        Update `ancestry`.
        """
        self.__ancestry = self._define_ancestry(x, self.laiobj.ancestry_map) if self.laiobj is not None else None

    @property
    def is_masked(self) -> bool:
        """
        Retrieve `is_masked`.
        
        Returns:
            **bool:** True if an ancestry file is passed for ancestry-specific masking, or False otherwise.
        """
        return self.__is_masked

    @is_masked.setter
    def is_masked(self, x: bool) -> None:
        """
        Update `is_masked`.
        """
        self.__is_masked = x

    @property
    def average_strands(self) -> bool:
        """
        Retrieve `average_strands`.
        
        Returns:
            **bool:** True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
        """
        return self.__average_strands

    @average_strands.setter
    def average_strands(self, x: bool) -> None:
        """
        Update `average_strands`.
        """
        self.__average_strands = x

    @property
    def force_nan_incomplete_strands(self) -> bool:
        """
        Retrieve `force_nan_incomplete_strands`.
        
        Returns:
            **bool**: If `True`, sets the result to NaN if either haplotype in a pair is NaN.
                      Otherwise, computes the mean while ignoring NaNs (e.g., 0|NaN -> 0, 1|NaN -> 1).
        """
        return self.__force_nan_incomplete_strands

    @force_nan_incomplete_strands.setter
    def force_nan_incomplete_strands(self, x: bool) -> None:
        """
        Update `force_nan_incomplete_strands`.
        """
        self.__force_nan_incomplete_strands = x

    @property
    def is_weighted(self) -> bool:
        """
        Retrieve `is_weighted`.
        
        Returns:
            **bool:** True if weights are provided in the labels file, or False otherwise.
        """
        return self.__is_weighted

    @is_weighted.setter
    def is_weighted(self, x: bool) -> None:
        """
        Update `is_weighted`.
        """
        self.__is_weighted = x

    @property
    def groups_to_remove(self) -> Dict[int, List[str]]:
        """
        Retrieve `groups_to_remove`.
        
        Returns:
            **dict of int to list of str:** Dictionary specifying groups to exclude from analysis. Keys are array numbers, and values are 
                lists of groups to remove for each array. Example: `{1: ['group1', 'group2'], 2: [], 3: ['group3']}`.
        """
        return self.__groups_to_remove

    @groups_to_remove.setter
    def groups_to_remove(self, x: Dict[int, List[str]]) -> None:
        """
        Update `groups_to_remove`.
        """
        self.__groups_to_remove = x

    @property
    def min_percent_snps(self) -> float:
        """
        Retrieve `min_percent_snps`.
        
        Returns:
            **float:** 
                Minimum percentage of SNPs to be known in an individual for an individual to be included in the analysis. 
                All individuals with fewer percent of unmasked SNPs than this threshold will be excluded.
        """
        return self.__min_percent_snps

    @min_percent_snps.setter
    def min_percent_snps(self, x: float) -> None:
        """
        Update `min_percent_snps`.
        """
        self.__min_percent_snps = x

    @property
    def group_snp_frequencies_only(self) -> bool:
        """
        Retrieve `group_snp_frequencies_only`.
        
        Returns:
            **bool:** 
                If True, maasMDS is performed exclusively on group-level SNP frequencies, ignoring individual-level data. This applies 
                when `is_weighted` is set to True and a `combination` column is provided in the `labels_file`,  meaning individuals are 
                aggregated into groups based on their assigned labels. If False, maasMDS is performed on individual-level SNP data alone 
                or on both individual-level and group-level SNP frequencies when `is_weighted` is True and a `combination` column is provided.
        """
        return self.__group_snp_frequencies_only

    @group_snp_frequencies_only.setter
    def group_snp_frequencies_only(self, x: bool) -> None:
        """
        Update `group_snp_frequencies_only`.
        """
        self.__group_snp_frequencies_only = x

    @property
    def save_masks(self) -> bool:
        """
        Retrieve `save_masks`.
        
        Returns:
            **bool:** True if the masked matrices are to be saved in a `.npz` file, or False otherwise.
        """
        return self.__save_masks

    @save_masks.setter
    def save_masks(self, x: bool) -> None:
        """
        Update `save_masks`.
        """
        self.__save_masks = x

    @property
    def load_masks(self) -> bool:
        """
        Retrieve `load_masks`.
        
        Returns:
            **bool:** 
                True if the masked matrices are to be loaded from a pre-existing `.npz` file specified 
                by `masks_file`, or False otherwise.
        """
        return self.__load_masks

    @load_masks.setter
    def load_masks(self, x: bool) -> None:
        """
        Update `load_masks`.
        """
        self.__load_masks = x

    @property
    def masks_file(self) -> Union[str, pathlib.Path]:
        """
        Retrieve `masks_file`.
        
        Returns:
            **str or pathlib.Path:** Path to the `.npz` file used for saving/loading masked matrices.
        """
        return self.__masks_file

    @masks_file.setter
    def masks_file(self, x: Union[str, pathlib.Path]) -> None:
        """
        Update `masks_file`.
        """
        self.__masks_file = x

    @property
    def distance_type(self) -> str:
        """
        Retrieve `distance_type`.
        
        Returns:
            **str:** 
                Distance metric to use. Options to choose from are: 'Manhattan', 'RMS' (Root Mean Square), 'AP' (Average Pairwise).
                If `average_strands=True`, use 'distance_type=AP'.
        """
        return self.__distance_type

    @distance_type.setter
    def distance_type(self, x: str) -> None:
        """
        Update `distance_type`.
        """
        self.__distance_type = x

    @property
    def n_components(self) -> int:
        """
        Retrieve `n_components`.
        
        Returns:
            **int:** The number of principal components.
        """
        return self.__n_components

    @n_components.setter
    def n_components(self, x: int) -> None:
        """
        Update `n_components`.
        """
        self.__n_components = x

    @property
    def rsid_or_chrompos(self) -> int:
        """
        Retrieve `rsid_or_chrompos`.
        
        Returns:
            **int:** Format indicator for SNP IDs in the SNP data. Use 1 for `rsID` format or 2 for `chromosome_position`.
        """
        return self.__rsid_or_chrompos

    @rsid_or_chrompos.setter
    def rsid_or_chrompos(self, x: int) -> None:
        """
        Update `rsid_or_chrompos`.
        """
        self.__rsid_or_chrompos = x

    @property
    def X_new_(self) -> Optional[np.ndarray]:
        """
        Retrieve `X_new_`.

        Returns:
            **array of shape (n_haplotypes_, n_components):** 
                The transformed SNP data projected onto the `n_components` principal components.
                n_haplotypes_ is the number of haplotypes, potentially reduced if filtering is applied 
                (`min_percent_snps > 0`). For diploid individuals without filtering, the shape is 
                `(n_samples * 2, n_components)`.
        """
        return self.__X_new_

    @X_new_.setter
    def X_new_(self, x: np.ndarray) -> None:
        """
        Update `X_new_`.
        """
        self.__X_new_ = x

    @property
    def haplotypes_(self) -> Optional[List[str]]:
        """
        Retrieve `haplotypes_`.

        Returns:
            list of str:
                A list of unique haplotype identifiers.
        """
        if isinstance(self.__haplotypes_, np.ndarray):
            return self.__haplotypes_.ravel().tolist()  # Flatten and convert NumPy array to a list
        elif isinstance(self.__haplotypes_, list):
            if len(self.__haplotypes_) == 1 and isinstance(self.__haplotypes_[0], np.ndarray):
                return self.__haplotypes_[0].ravel().tolist()  # Handle list containing a single array
            return self.__haplotypes_  # Already a flat list
        elif self.__haplotypes_ is None:
            return None  # If no haplotypes are set
        else:
            raise TypeError("`haplotypes_` must be a list or a NumPy array.")

    @haplotypes_.setter
    def haplotypes_(self, x: Union[np.ndarray, List[str]]) -> None:
        """
        Update `haplotypes_`.
        """
        if isinstance(x, np.ndarray):
            self.__haplotypes_ = x.ravel().tolist()  # Flatten and convert to a list
        elif isinstance(x, list):
            if len(x) == 1 and isinstance(x[0], np.ndarray):  # Handle list containing a single array
                self.__haplotypes_ = x[0].ravel().tolist()
            else:
                self.__haplotypes_ = x  # Use directly if already a list
        else:
            raise TypeError("`x` must be a list or a NumPy array.")

    @property
    def samples_(self) -> Optional[List[str]]:
        """
        Retrieve `samples_`.

        Returns:
            list of str:
                A list of sample identifiers based on `haplotypes_` and `average_strands`.
        """
        haplotypes = self.haplotypes_
        if haplotypes is None:
            return None
        if self.__average_strands:
            return haplotypes
        else:
            return [x[:-2] for x in haplotypes]

    @property
    def variants_id_(self) -> Optional[np.ndarray]:
        """
        Retrieve `variants_id_`.

        Returns:
            **array of shape (n_snp,):** 
                An array containing unique identifiers (IDs) for each SNP,
                potentially reduced if there are SNPs not present in the `laiobj`.
                The format will depend on `rsid_or_chrompos`.
        """
        return self.__variants_id_

    @variants_id_.setter
    def variants_id_(self, x: np.ndarray) -> None:
        """
        Update `variants_id_`.
        """
        self.__variants_id_ = x

    @property
    def n_haplotypes(self) -> Optional[int]:
        """
        Retrieve `n_haplotypes`.

        Returns:
            **int:**
                The total number of haplotypes, potentially reduced if filtering is applied 
                (`min_percent_snps > 0`).
        """
        return len(self.__haplotypes_)

    @property
    def n_samples(self) -> Optional[int]:
        """
        Retrieve `n_samples`.

        Returns:
            **int:**
                The total number of samples, potentially reduced if filtering is applied 
                (`min_percent_snps > 0`).
        """
        return len(np.unique(self.samples_))

    @staticmethod
    def _define_ancestry(ancestry, ancestry_map):
        """
        Determine the ancestry index based on different input types.

        Args:
            ancestry (int or str): The ancestry input, which can be:
                - An integer (e.g., 0, 1, 2).
                - A string representation of an integer (e.g., '0', '1').
                - A string matching one of the ancestry map values (e.g., 'Africa').
            ancestry_map (dict): A dictionary mapping ancestry indices (as strings) to ancestry names.

        Returns:
            int: The corresponding ancestry index.
        """
        if isinstance(ancestry, int):  
            return ancestry  
        elif isinstance(ancestry, str) and ancestry.isdigit():  
            return int(ancestry)  
        elif ancestry in ancestry_map.values():  
            return int(next(key for key, value in ancestry_map.items() if value == ancestry))  
        else:  
            raise ValueError(f"Invalid ancestry input: {ancestry}")

    @staticmethod
    def _load_masks_file(masks_file):
        mask_files = np.load(masks_file, allow_pickle=True)
        mask = mask_files['mask']
        rs_ID_list = mask_files['rs_ID_list']
        ind_ID_list = mask_files['ind_ID_list']
        groups = mask_files['labels']
        weights = mask_files['weights']
        return mask, rs_ID_list, ind_ID_list, groups, weights

    def fit_transform(
            self,
            snpobj: Optional['SNPObject'] = None, 
            laiobj: Optional['LocalAncestryObject'] = None,
            labels_file: Optional[str] = None,
            ancestry: Optional[str] = None,
            average_strands: Optional[bool] = None
        ) -> np.ndarray:
        """
        Fit the model to the SNP data stored in the provided `snpobj` and apply the dimensionality reduction on the same SNP data.

        Args:
            snpobj (SNPObject, optional): 
                A SNPObject instance.
            laiobj (LAIObject, optional): 
                A LAIObject instance.
            labels_file (str, optional): 
                Path to the labels file in .tsv format. The first column, `indID`, contains the individual identifiers, and the second 
                column, `label`, specifies the groups for all individuals. If `is_weighted=True`, a `weight` column with individual 
                weights is required. Optionally, `combination` and `combination_weight` columns can specify sets of individuals to be 
                combined into groups, with respective weights.
            ancestry (str, optional): 
                Ancestry for which dimensionality reduction is to be performed. Ancestry counter starts at 0.
            average_strands (bool, optional): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
                If None, defaults to `self.average_strands`.

        Returns:
            **array of shape (n_samples, n_components):** 
                The transformed SNP data projected onto the `n_components` principal components, stored in `self.X_new_`.
        """
        if snpobj is None:
            snpobj = self.snpobj
        if laiobj is None:
            laiobj = self.laiobj
        if labels_file is None:
            labels_file = self.labels_file
        if ancestry is None:
            ancestry = self.ancestry
        if average_strands is None:
            average_strands = self.average_strands
        
        if not self.is_masked:
            self.ancestry = 1
        if self.load_masks:
            # Load precomputed ancestry-based masked genotype matrix, SNP identifiers, haplotype identifiers, and weights
            mask, variants_id, haplotypes, _, weights = self._load_masks_file(self.masks_file)
        else:
            # Process genotype data with optional ancestry-based masking and return the corresponding SNP and individual identifiers
            mask, variants_id, haplotypes = process_calldata_gt(
                self.snpobj,
                self.laiobj,
                self.ancestry,
                self.average_strands,
                self.force_nan_incomplete_strands,
                self.is_masked, 
                self.rsid_or_chrompos
            )

            # Process individual genomic labels and weights, aligning them with a masked genotype matrix by 
            # filtering out low-coverage individuals, reordering data to match the matrix structure, and 
            # handling group-based adjustments
            mask, haplotypes, groups, weights = process_labels_weights(
                self.labels_file,
                mask,
                variants_id,
                haplotypes,
                self.average_strands,
                self.ancestry,
                self.min_percent_snps,
                self.group_snp_frequencies_only,
                self.groups_to_remove,
                self.is_weighted,
                self.save_masks,
                self.masks_file
            )
        
        distance_list = [[distance_mat(first=mask[self.ancestry], dist_func=self.distance_type)]]
        
        self.X_new_ = mds_transform(distance_list, groups, weights, haplotypes, self.n_components)
        
        self.haplotypes_ = haplotypes
        self.variants_id_ = variants_id



import abc
from pathlib import Path
from typing import Union, Optional
import logging
import numpy as np

log = logging.getLogger(__name__)


class WideBaseReader(abc.ABC):
    """
    Abstract class for global ancestry readers.
    """
    def __init__(
        self,
        Q_file: Union[str, Path],
        P_file: Optional[Union[str, Path]] = None,
        sample_file: Optional[Union[str, Path]] = None,
        snp_file: Optional[Union[str, Path]] = None,
        ancestry_file: Optional[Union[str, Path]] = None
    ) -> None:
        """
        Args:
            Q_file (str or pathlib.Path):
                Path to the file containing the Q matrix (per-sample ancestry proportions).
                It should end with `.Q` or `.txt`.
                The file should use space (' ') as the delimiter.
            P_file (str or pathlib.Path, optional):
                Path to the file containing the P/F matrix (per-ancestry SNP frequencies).
                It should end with `.P` or `.txt`.
                The file should use space (' ') as the delimiter. If None, P is not loaded.
            sample_file (str or pathlib.Path, optional):
                Path to the single-column file containing sample identifiers. 
                It should end with `.fam` or `.txt`.
                If None, sample identifiers are not loaded.
            snp_file (str or pathlib.Path, optional):
                Path to the single-column file containing SNP identifiers. 
                It should end with `.bim` or `.txt`.
                If None, SNP identifiers are not loaded.
            ancestry_file (str or pathlib.Path, optional):
                Path to the single-column file containing ancestry labels for each sample.
                It should end with `.map` or `.txt`.
                If None, ancestries are not loaded.
        """
        self.__Q_file = Path(Q_file)
        self.__P_file = Path(P_file) if P_file is not None else None
        self.__sample_file = Path(sample_file) if sample_file is not None else None
        self.__snp_file = Path(snp_file) if snp_file is not None else None
        self.__ancestry_file = Path(ancestry_file) if ancestry_file is not None else None

    @property
    def Q_file(self) -> Path:
        """
        Retrieve `Q_file`.

        Returns:
            **pathlib.Path:** 
                Path to the file containing the Q matrix (per-sample ancestry proportions).
                It should end with `.Q` or `.txt`.
                The file should use space (' ') as the delimiter.
        """
        return self.__Q_file

    @property
    def P_file(self) -> Optional[Path]:
        """
        Retrieve `P_file`.

        Returns:
            **pathlib.Path or None:** 
                Path to the file containing the P/F matrix (per-ancestry SNP frequencies).
                It should end with `.P` or `.txt`.
                The file should use space (' ') as the delimiter. If None, P is not loaded.
        """
        return self.__P_file

    @property
    def sample_file(self) -> Optional[Path]:
        """
        Retrieve `sample_file`.

        Returns:
            **pathlib.Path:** 
                Path to the single-column file containing sample identifiers. 
                It should end with `.fam` or `.txt`.
                If None, sample identifiers are not loaded.
        """
        return self.__sample_file
    
    @property
    def snp_file(self) -> Optional[Path]:
        """
        Retrieve `snp_file`.

        Returns:
            **pathlib.Path:** 
                Path to the single-column file containing SNP identifiers. 
                It should end with `.bim` or `.txt`.
                If None, SNP identifiers are not loaded.
        """
        return self.__snp_file

    @property
    def ancestry_file(self) -> Optional[Path]:
        """
        Retrieve `ancestry_file`.

        Returns:
            **pathlib.Path:** 
                Path to the single-column file containing ancestry labels for each sample.
                It should end with `.map` or `.txt`.
                If None, ancestries are not loaded.
        """
        return self.__ancestry_file

    def _read_sample_ids(self) -> Optional[np.ndarray]:
        """
        Read sample identifiers from `sample_file` file.
        """
        if self.sample_file is None:
            return None
        
        log.info(f"Reading sample identifiers from '{self.sample_file}'...")
        if self.sample_file.suffix == ".txt":
            return np.genfromtxt(self.sample_file, dtype=str)
        elif self.sample_file.suffix == ".fam":
            return np.genfromtxt(self.sample_file, dtype=str, usecols=1)
        else:
            raise ValueError(
                f"Unsupported file extension for sample identifiers: {self.sample_file.suffix}"
                "Supported extensions are: .txt, .fam."
            )
    
    def _read_snps(self) -> Optional[np.ndarray]:
        """
        Read SNP identifiers from `snp_file` file.
        """
        if self.snp_file is None:
            return None
        
        log.info(f"Reading SNP identifiers from '{self.snp_file}'...")
        if self.snp_file.suffix == ".txt":
            return np.genfromtxt(self.snp_file, dtype=str)
        elif self.snp_file.suffix == ".bim":
            return np.genfromtxt(self.snp_file, dtype=str, usecols=1)
        else:
            raise ValueError(
                f"Unsupported file extension for SNP identifiers: {self.snp_file.suffix}"
                "Supported extensions are: .txt, .bim."
            )

    def _read_ancestries(self) -> Optional[np.ndarray]:
        """
        Read ancestries for each sample from `ancestry_file`.
        """
        if self.ancestry_file is None:
            return None
        
        log.info(f"Reading ancestries for each sample from '{self.ancestry_file}'...")
        if self.ancestry_file.suffix in [".map", ".txt"]:
            return np.genfromtxt(self.ancestry_file, dtype=str)
        else:
            raise ValueError(
                f"Unsupported file extension for ancestries: {self.ancestry_file.suffix}"
                "Supported extension: .map."
            )

    @abc.abstractmethod
    def read(self) -> None:
        """
        Abstract method to read data from the provided files.

        Subclasses must implement this method to read and parse the data from
        the Q and P files, as well as sample identifiers, SNPs, and ancestries if provided.
        The implementation should construct an instance of `snputils.ancestry.genobj.GlobalAncestryObject` 
        based on the read data.
        """
        pass



import pytest
import numpy as np
from .utils import create_benchmark_test


def read_bed_snputils(path):
    """Read BED fileset using snputils"""
    import snputils
    return snputils.read_bed(path, sum_strands=True, fields=["GT"]).calldata_gt


def read_bed_pgenlib(path):
    """Read BED fileset using Pgenlib"""
    import pgenlib
    with open(path + '.fam', 'r') as f:
        sample_ct = sum(1 for _ in f)  # Get sample count from fam file
    pgen = pgenlib.PgenReader(str.encode(path + '.bed'), raw_sample_ct=sample_ct)
    variant_ct = pgen.get_variant_ct()
    genotypes = np.empty((variant_ct, sample_ct), dtype=np.int8)
    variant_idxs = np.arange(variant_ct, dtype=np.uint32)
    pgen.read_list(variant_idxs, genotypes)
    pgen.close()
    return genotypes


def read_bed_hail(path):
    """Read BED fileset using hail"""
    import hail as hl
    hl.init()
    mt = hl.import_plink(path + '.bed', path + '.bim', path + '.fam')
    n_variants = mt.count_rows()
    n_samples = mt.count_cols()
    mt = np.array(mt.GT.n_alt_alleles().collect(), dtype=np.uint8).reshape((n_variants, n_samples))
    hl.stop()
    return mt


def read_bed_sgkit(path):
    """Read BED fileset using sgkit"""
    from sgkit.io import plink
    return np.sum(plink.read_plink(path=path).call_genotype.to_numpy() == 0, axis=2, dtype=np.uint8)


def read_bed_pandas_plink(path):
    """Read BED fileset using pandas-plink"""
    import pandas_plink
    _, _, genotypes = pandas_plink.read_plink(path)
    return 2 - genotypes.compute().astype(np.uint8)


def read_bed_plinkio(path):
    """Read BED fileset using plinkio"""
    from plinkio import plinkfile
    return np.array([2 - np.array(row) for row in plinkfile.open(path)], dtype=np.uint8)


def read_bed_pysnptools(path):
    """Read BED fileset using pysnptools"""
    from pysnptools.snpreader import Bed
    return (2 - Bed(path).read().val.T.astype(np.uint8))


READERS = [
    (read_bed_snputils, "snputils"),
    (read_bed_pgenlib, "pgenlib"),
    (read_bed_hail, "hail"),
    (read_bed_sgkit, "sgkit"),
    (read_bed_pandas_plink, "pandas-plink"),
    (read_bed_plinkio, "plinkio"),
    (read_bed_pysnptools, "pysnptools"),
]


@pytest.mark.benchmark(group="BED-readers", warmup=False, min_rounds=3)
@pytest.mark.parametrize("reader,name", READERS)
def test_bed_readers(benchmark, reader, name, path, memory_profile):
    """Benchmark readers and verify output"""
    ref_array = read_bed_snputils(path)
    create_benchmark_test(benchmark, reader, path, name, ref_array, memory_profile)



import logging
from pathlib import Path
from typing import Union, Optional
import numpy as np

from .base import WideBaseWriter
from snputils.ancestry.genobj.wide import GlobalAncestryObject

log = logging.getLogger(__name__)


class AdmixtureWriter(WideBaseWriter):
    """
    A writer class for exporting global ancestry data from a 
    `snputils.ancestry.genobj.GlobalAncestryObject` into multiple ADMIXTURE files.
    """
    def __init__(
        self, 
        wideobj: GlobalAncestryObject, 
        file_prefix: Union[str, Path]
    ) -> None:
        """
        Args:
            wideobj (GlobalAncestryObject): 
                A GlobalAncestryObject instance.
            file_prefix (str or pathlib.Path): 
                Prefix for output file names, including directory path but excluding file extensions. 
                The prefix is used to generate specific file names for each output, with file-specific 
                suffixes appended as described above (e.g., `file_prefix.n_ancestries.Q` for the Q matrix file).
        """
        super(AdmixtureWriter, self).__init__(wideobj, file_prefix)
        self.__Q_file = self.file_prefix.with_suffix(f".{self.wideobj.n_ancestries}.Q")
        self.__P_file = self.file_prefix.with_suffix(f".{self.wideobj.n_ancestries}.P")

        self.__sample_file = self.file_prefix.with_suffix(".sample_ids.txt") if self.wideobj.samples is not None else None
        self.__snp_file = self.file_prefix.with_suffix(".snp_ids.txt") if self.wideobj.snps is not None else None
        self.__ancestry_file = self.file_prefix.with_suffix(".map") if self.wideobj.ancestries is not None else None

    @property
    def wideobj(self) -> GlobalAncestryObject:
        """
        Retrieve `wideobj`.

        Returns:
            **GlobalAncestryObject:** A GlobalAncestryObject instance.
        """
        return self.__wideobj

    @property
    def file_prefix(self) -> Path:
        """
        Retrieve `file_prefix`.

        Returns:
            **pathlib.Path:** 
                Prefix for output file names, including directory path but excluding file extensions. 
                The prefix is used to generate specific file names for each output, with file-specific 
                suffixes appended as described above (e.g., `file_prefix.n_ancestries.Q` for the Q matrix file).
        """
        return self.__file_prefix

    @property
    def Q_file(self) -> Path:
        """
        Retrieve `Q_file`.

        Returns:
            **pathlib.Path:** 
                Path to the `.Q` file that will store the Q matrix (per-sample ancestry proportions).
        """
        return self.__Q_file
    
    @property
    def P_file(self) -> Path:
        """
        Retrieve `P_file`.

        Returns:
            **pathlib.Path:** 
                Path to the `.P` file that will store the P/F matrix (per-ancestry SNP frequencies).
        """
        return self.__P_file
    
    @property
    def sample_file(self) -> Optional[Path]:
        """
        Retrieve `sample_file`.

        Returns:
            **pathlib.Path:** 
                Path to the `.txt` the file that will store sample identifiers. 
                If None, sample identifiers are not saved.
        """
        return self.__sample_file
    
    @property
    def snp_file(self) -> Optional[Path]:
        """
        Retrieve `snp_file`.

        Returns:
            **pathlib.Path:** 
                Path to the `.txt` file that will store SNP identifiers. 
                If None, SNP identifiers are not saved.
        """
        return self.__snp_file
    
    @property
    def ancestry_file(self) -> Optional[Path]:
        """
        Retrieve `ancestry_file`.

        Returns:
            **pathlib.Path:** 
                Path to the `.map` file that will store ancestry labels for each sample. 
                If None, ancestries are not saved.
        """
        return self.__ancestry_file

    def _write_Q(self):
        log.info(f"Writing Q matrix to '{self.Q_file}'...")
        np.savetxt(self.Q_file, self.wideobj.Q, delimiter=" ")
        log.info(f"Finished writing Q matrix to '{self.Q_file}'.")

    def _write_P(self):
        log.info(f"Writing P matrix to '{self.P_file}'...")
        np.savetxt(self.P_file, self.wideobj.P, delimiter=" ")
        log.info(f"Finished writing P matrix to '{self.P_file}'.")

    def _write_sample_ids(self):
        if self.wideobj.samples is not None:
            log.info(f"Writing sample IDs to '{self.sample_file}'...")
            np.savetxt(self.sample_file, self.wideobj.samples, fmt="%s")
            log.info(f"Finished writing sample IDs to '{self.sample_file}'.")

    def _write_snps(self):
        if self.wideobj.snps is not None:
            log.info(f"Writing SNP IDs to '{self.snp_file}'...")
            np.savetxt(self.snp_file, self.wideobj.snps, fmt="%s")
            log.info(f"Finished writing SNP IDs to '{self.snp_file}'.")

    def _write_ancestries(self):
        if self.wideobj.ancestries is not None:
            log.info(f"Writing ancestry information to '{self.ancestry_file}'...")
            np.savetxt(self.ancestry_file, self.wideobj.ancestries, fmt="%s")
            log.info(f"Finished writing ancestry information to '{self.ancestry_file}'.")

    def write(self) -> None:
        """
        Write the data contained in the `wideobj` instance into the multiple ADMIXTURE files
        with the specified `file_prefix`. If the files already exist, they will be overwritten.

        **Output files:**

        - `<file_prefix>.K.Q`: Q matrix file. The file uses space (' ') as the delimiter.
        - `<file_prefix>.K.P`: P matrix file. The file uses space (' ') as the delimiter.
        - `<file_prefix>.sample_ids.txt`: Sample IDs file (if sample IDs are available).
        - `<file_prefix>.snp_ids.txt`: SNP IDs file (if SNP IDs are available).
        - `<file_prefix>.map`: Ancestry file (if ancestries information is available).

        where `K` is the total number of ancestries.
        """
        log.info(f"Preparing to write ADMIXTURE files with prefix '{self.file_prefix}'...")
        
        self.file_prefix.parent.mkdir(parents=True, exist_ok=True)
        
        self._write_Q()
        self._write_P()
        self._write_sample_ids()
        self._write_snps()
        self._write_ancestries()

        log.info(f"Finished writing all ADMIXTURE files with prefix '{self.file_prefix}'.")

WideBaseWriter.register(AdmixtureWriter)



import logging
from typing import Optional, List
from pathlib import Path
import gzip
import csv

import allel
import numpy as np
import polars as pl

from snputils.snp.genobj.snpobj import SNPObject
from snputils.snp.io.read.base import SNPBaseReader

log = logging.getLogger(__name__)


@SNPBaseReader.register
class VCFReader(SNPBaseReader):
    def read(
        self,
        fields: Optional[List[str]] = None,
        exclude_fields: Optional[List[str]] = None,
        rename_fields: Optional[dict] = None,
        fills: Optional[dict] = None,
        region: Optional[str] = None,
        samples: Optional[List[str]] = None,
        sum_strands: bool = False,
    ) -> SNPObject:
        """
        Read a vcf file into a SNPObject.

        Args:
            fields: Fields to extract data for. e.g., ['variants/CHROM', 'variants/POS',
                'calldata/GT']. If you are feeling lazy, you can drop the 'variants/'
                and 'calldata/' prefixes, in which case the fields will be matched
                against fields declared in the VCF header, with variants taking priority
                over calldata if a field with the same ID exists both in INFO and FORMAT
                headers. I.e., ['CHROM', 'POS', 'DP', 'GT'] will work, although watch out
                for fields like 'DP' which can be both INFO and FORMAT. To extract all
                fields, provide just the string '*'. To extract all variants fields
                (including all INFO fields) provide 'variants/*'. To extract all
                calldata fields (i.e., defined in FORMAT headers) provide 'calldata/*'.
            exclude_fields: Fields to exclude. E.g., for use in combination with fields='*'.
            rename_fields: Fields to be renamed. Should be a dictionary mapping old to new names.
            fills: Override the fill value used for empty values. Should be a dictionary
                mapping field names to fill values.
            region: Genomic region to extract variants for. If provided, should be a
                tabix-style region string, which can be either just a chromosome name
                (e.g., '2L'), or a chromosome name followed by 1-based beginning and
                end coordinates (e.g., '2L:100000-200000'). Note that only variants
                whose start position (POS) is within the requested range will be included.
                This is slightly different from the default tabix behaviour, where a
                variant (e.g., deletion) may be included if its position (POS) occurs
                before the requested region but its reference allele overlaps the
                region - such a variant will not be included in the data returned
                by this function.
            samples: Selection of samples to extract calldata for. If provided, should be
                a list of strings giving sample identifiers. May also be a list of
                integers giving indices of selected samples.
            sum_strands: If True, maternal and paternal strands are combined into a single `int8` array with values `{0, 1, 2`}. 
                If False, strands are stored separately as an `int8` array with values `{0, 1}` for each strand.

        Returns:
            **SNPObject**: 
                A SNPObject instance.
        """
        log.info(f"Reading {self.filename}")

        vcf_dict = allel.read_vcf(
            str(self.filename),
            fields=fields,
            exclude_fields=exclude_fields,
            rename_fields=rename_fields,
            fills=fills,
            region=region,
            samples=samples,
            alt_number=1,
        )
        assert vcf_dict is not None  # suppress Flake8 warning

        genotypes = vcf_dict["calldata/GT"].astype(np.int8)
        if sum_strands:
            genotypes = genotypes.sum(axis=2, dtype=np.int8)

        snpobj = SNPObject(
            calldata_gt=genotypes,
            samples=vcf_dict["samples"],
            variants_ref=vcf_dict["variants/REF"],
            variants_alt=vcf_dict["variants/ALT"],
            variants_chrom=vcf_dict["variants/CHROM"],
            variants_filter_pass=vcf_dict["variants/FILTER_PASS"],
            variants_id=vcf_dict["variants/ID"],
            variants_pos=vcf_dict["variants/POS"],
            variants_qual=vcf_dict["variants/QUAL"],
        )

        log.info(f"Finished reading {self.filename}")
        return snpobj


def _get_vcf_col_names_and_sep(vcf_path: str, separator: Optional[str] = None):
    """
    Get the column names and separator used in the VCF file.

    Args:
        vcf_path: The path to the VCF file.
        separator: Separator character. If None, the separator is automatically detected.

    Returns:
        col_names: List of column names.
        separator: Separator character.
    """
    vcf_path = Path(vcf_path)
    if vcf_path.suffixes[-2:] == ['.vcf', '.gz']:
        open_func = gzip.open
        mode = 'rt'
    elif vcf_path.suffix == '.vcf':
        open_func = open
        mode = 'r'
    else:
        raise ValueError(f"Unsupported file extension: {vcf_path.suffixes}")

    with open_func(vcf_path, mode) as file:
        for line in file:
            if line.startswith("#CHROM"):
                if separator is None:
                    separator = csv.Sniffer().sniff(file.readline()).delimiter
                col_names = [x.strip() for x in line.split(separator)]
                break

    return col_names, separator


def _infer_col_data_types(names: List):
    """
    Infer data types for VCF columns.

    Args:
        names: List of column names.

    Returns:
        col_dtypes: Dictionary mapping column names to data types.
    """
    col_dtypes = {name: pl.Utf8 for name in names}
    col_dtypes['POS'] = pl.Int32
    col_dtypes['#CHROM'] = pl.String

    return col_dtypes


def _extract_columns(names: List[str], fields: List[str], exclude_fields: List[str],
                     samples: List[str]) -> List[str]:
    """
    Extracts columns based on specified `fields`, `exclude_fields` and `samples`.

    Args:
        names: List of column names.
        fields: Fields to extract data for. This parameter specifies which data fields
            from the VCF file should be included in the result. To extract all fields,
            provide just the string '*'.
        exclude_fields: Fields to exclude. E.g., for use in combination with fields='*'.
        samples: Selection of samples to extract calldata for. If provided, should be
            a list of strings giving sample identifiers. May also be a list of
            integers giving indices of selected samples.

    Returns:
        field_columns: List of field columns.
        sample_columns: List of sample columns.
        selected_column_idxs: List of selected column indices.
    """

    # Define standard field names in a VCF file
    field_names = ['#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT']

    # Find the index of the first column that is not a standard field name
    first_sample_idx = next(i for i, col in enumerate(names) if col not in field_names)

    # Identify field columns as all columns before the first sample column
    field_columns = names[:first_sample_idx]

    if fields != '*' and fields is not None:
        # Filter field columns to contain those in `fields`
        field_columns = list(set(field_columns).intersection(set(fields)))
    elif fields == '*' and exclude_fields is not None:
        field_columns = list(set(field_columns) - set(exclude_fields))

    # Sample columns are all columns starting from the first sample column
    sample_columns = names[first_sample_idx:]

    if samples is not None:
        if len(samples) == 0:
            sample_columns = []
        elif type(samples[0]) is int:
            sample_columns = list(np.array(sample_columns)[samples])
        else:
            sample_columns = list(set(sample_columns).intersection(set(samples)))

    # Create a dictionary mapping column names to their indices
    column_idx_map = {name: index for index, name in enumerate(names)}

    # Create a list of selected column indices
    selected_column_idxs = [column_idx_map[col] for col in field_columns + sample_columns]

    # Sort the selected column indices
    selected_column_idxs.sort()

    return field_columns, sample_columns, selected_column_idxs


@SNPBaseReader.register
class VCFReaderPolars(SNPBaseReader):
    """Reads a VCF file and processes it into a SNPObject."""

    def __init__(self, filename: str):
        self._filename = filename

    def read(self,
             fields: Optional[List[str]] = None,
             exclude_fields: Optional[List[str]] = None,
             region: Optional[str] = None,
             samples: Optional[List[str]] = None,
             sum_strands: Optional[bool] = False,
             separator: Optional[str] = None
             ) -> SNPObject:
        """
        Read a vcf file into a SNPObject.

        Args:
            fields: Fields to extract data for. This parameter specifies which data fields
                from the VCF file should be included in the result. Available options include
                '#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', and 'FORMAT'.
                To extract all fields, provide just the string '*' or the default None.
            exclude_fields: Fields to exclude for use in combination with fields='*'.
                Available options include '#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER',
                'INFO', and 'FORMAT'.
            region: Genomic region to extract variants for. If provided, it should be a
                tabix-style region string, specifying a chromosome name and optionally
                beginning and end coordinates (e.g., '2L:100000-200000'). TODO
            samples: Selection of samples to extract calldata for. If provided, should be
                a list of strings giving sample identifiers. May also be a list of
                integers giving indices of selected samples.  If an empty list is provided,
                no samples are extracted.
            sum_strands: True if the maternal and paternal strands are to be summed together,
                False if the strands are to be stored separately.
            separator: Separator used in the pvar file. If None, the separator is automatically detected.
                If the automatic detection fails, please specify the separator manually.

        Returns:
            snpobj: SNPObject containing the data from the VCF file. The format and content
                of this object depend on the specified parameters and the content of the VCF file.
        """
        # TODO: add support for excluding GT

        log.info(f"Reading {self._filename}")

        try:
            # Get the names and data types for the VCF file
            col_names, separator = _get_vcf_col_names_and_sep(self._filename)
            col_dtypes = _infer_col_data_types(col_names)

            # Extract columns to read based on specified 'fields', 'exclude_fields' and 'samples'
            # `selected_column_idxs` contains the indexes of the selected columns
            fields, samples, selected_column_idxs = _extract_columns(col_names, fields, exclude_fields, samples)

            # Read the VCF file into a Polars DataFrame
            vcf = pl.read_csv(
                self._filename,
                comment_prefix='#',
                has_header=False,
                separator=separator,
                columns=selected_column_idxs,
                dtypes=col_dtypes
            )

            log.debug("vcf polars read")

            if not samples:
                genotypes = np.array([])
            else:  # note that if samples is None, all samples are read
                # Process maternal strand:
                # Extract the first position from genotype, e.g., 0|1 -> 0
                # Replace missing values codified as ".", "-", or "" with -1, necessary for integer casting
                genotype_maternal = vcf[samples].select(pl.all().str.slice(0, length=1)) \
                                                .select(pl.all().replace({
                                                    ".": -1,
                                                    "-": -1,
                                                    "": -1
                                                })).cast(pl.Int8)

                # Process paternal strand:
                # Extract the third position from genotype, e.g., 0|1 -> 1
                # Convert ":" to ".." such that if only maternal strand is present, paternal strand is set to missing,
                # e.g., 0:0.982 -> 0..0.982 -> . -> -1
                # Replace missing values with -1
                genotype_paternal = vcf[samples].select(pl.all().str.replace_all('-1', '.')) \
                                                .select(pl.all().str.replace(':', '..')) \
                                                .select(pl.all().str.slice(2, length=1)) \
                                                .select(pl.all().replace({
                                                    ".": -1,
                                                    "": -1
                                                })).cast(pl.Int8)

                # Combine maternal and paternal genotypes
                genotypes = np.dstack((genotype_maternal, genotype_paternal))

                if sum_strands:
                    genotypes = genotypes.sum(axis=2, dtype=np.int8)

            # Create a SNPObject with the processed data
            snpobj = SNPObject(
                calldata_gt=genotypes,
                samples=np.array(samples),
                variants_ref=vcf['REF'].to_numpy() if 'REF' in fields else np.array([]),
                variants_alt=vcf['ALT'].to_numpy() if 'ALT' in fields else np.array([]),
                variants_chrom=vcf['#CHROM'].to_numpy() if '#CHROM' in fields else np.array([]),
                variants_filter_pass=vcf['FILTER'].to_numpy() if 'FILTER' in fields else np.array([]),
                variants_id=vcf['ID'].to_numpy() if 'ID' in fields else np.array([]),
                variants_pos=vcf['POS'].to_numpy() if 'POS' in fields else np.array([]),
                variants_qual=vcf['QUAL'].to_numpy() if 'QUAL' in fields else np.array([])
            )

            log.info(f"Finished reading {self.filename}")

            return snpobj

        except Exception as e:
            print(f"An error occurred: {e}.")
            from snputils.snp.io.read import VCFReader

            # Instantiate a VCFReader object and read SNP data
            reader = VCFReader(self._filename)
            snpobj = reader.read()

            return snpobj



import pathlib
import os
import time
import gc
import logging
import logging.config
import numpy as np
import copy
from typing import Optional, Dict, List, Union
from sklearn.decomposition import TruncatedSVD

from snputils.snp.genobj.snpobj import SNPObject
from snputils.ancestry.genobj.local import LocalAncestryObject
from ._utils.gen_tools import process_calldata_gt, process_labels_weights
from ._utils.iterative_svd import IterativeSVD


class mdPCA:
    """
    A class for performing missing data principal component analysis (mdPCA) on SNP data.

    The mdPCA class focuses on genotype segments from the ancestry of interest when the `is_masked` flag is set to `True`. It offers 
    flexible processing options, allowing either separate handling of masked haplotype strands or combining (averaging) strands into a 
    single composite representation for each individual. Moreover, the analysis can be performed on individual-level data, group-level SNP 
    frequencies, or a combination of both.

    If the `snpobj`, `laiobj`, `labels_file`, and `ancestry` parameters are all provided during instantiation, the `fit_transform` method 
    will be automatically called, applying the specified mdPCA method to transform the data upon instantiation.
    """
    def __init__(
        self,
        method: str = 'weighted_cov_pca',
        snpobj: Optional['SNPObject'] = None,
        laiobj: Optional['LocalAncestryObject'] = None,
        labels_file: Optional[str] = None,
        ancestry: Optional[Union[int, str]] = None,
        is_masked: bool = True,
        average_strands: bool = False,
        force_nan_incomplete_strands: bool = False,
        is_weighted: bool = False,
        groups_to_remove: List[str] = None,
        min_percent_snps: float = 4,
        group_snp_frequencies_only: bool = True,
        save_masks: bool = False,
        load_masks: bool = False,
        masks_file: Union[str, pathlib.Path] = 'masks.npz',
        output_file: Union[str, pathlib.Path] = 'output.tsv',
        covariance_matrix_file: Optional[str] = None,
        n_components: int = 2,
        rsid_or_chrompos: int = 2,
        percent_vals_masked: float = 0
    ):
        """
        Args:
            method (str, default='weighted_cov_pca'): 
                The PCA method to use for dimensionality reduction. Options include:
                - `'weighted_cov_pca'`: 
                    Simple covariance-based PCA, weighted by sample strengths.
                - `'regularized_optimization_ils'`: 
                    Regularized optimization followed by iterative, weighted (via the strengths) least squares projection of 
                    missing samples using the original covariance matrix (considering only relevant elements not missing in 
                    the original covariance matrix for those samples).
                - `'cov_matrix_imputation'`: 
                    Eigen-decomposition of the covariance matrix after first imputing the covariance matrix missing values 
                    using the Iterative SVD imputation method.
                - `'cov_matrix_imputation_ils'`: 
                    The method of 'cov_matrix_imputation', but where afterwards missing samples are re-projected onto the space 
                    given by 'cov_matrix_imputation' using the same iterative method on the original covariance matrix just 
                    as done in 'regularized_optimization_ils'.
                - `'nonmissing_pca_ils'`: 
                    The method of 'weighted_cov_pca' on the non-missing samples, followed by the projection of missing samples onto 
                    the space given by 'weighted_cov_pca' using the same iterative method on the original covariance matrix just as 
                    done in 'regularized_optimization_ils'.
            snpobj (SNPObject, optional): 
                A SNPObject instance.
            laiobj (LAIObject, optional): 
                A LAIObject instance.
            labels_file (str, optional): 
                Path to a `.tsv` file with metadata on individuals, including population labels, optional weights, and groupings. 
                The `indID` column must contain unique individual identifiers matching those in `laiobj` and `snpobj` for proper alignment. 
                The `label` column assigns population groups. If `is_weighted=True`, a `weight` column must be provided, assigning a weight to 
                each individual, where those with a weight of zero are removed. Optional columns include `combination` and `combination_weight` 
                to aggregate individuals into combined groups, where SNP frequencies represent their sequences. The `combination` column assigns 
                each individual to a specific group (0 for no combination, 1 for the first group, 2 for the second, etc.). All members of a group 
                must share the same `label` and `combination_weight`. If `combination_weight` column is not provided, the combinations are 
                assigned a default weight of `1`. Individuals excluded via `groups_to_remove` or those falling below `min_percent_snps` are removed 
                from the analysis.
            ancestry (int or str, optional): 
                Ancestry for which dimensionality reduction is to be performed. Ancestry counter starts at `0`. The ancestry input can be:
                - An integer (e.g., 0, 1, 2).
                - A string representation of an integer (e.g., '0', '1').
                - A string matching one of the ancestry map values (e.g., 'Africa').
            is_masked (bool, default=True): 
                If `True`, applies ancestry-specific masking to the genotype matrix, retaining only genotype data 
                corresponding to the specified `ancestry`. If `False`, uses the full, unmasked genotype matrix.
            average_strands (bool, default=False): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
            force_nan_incomplete_strands (bool): 
                If `True`, sets the result to NaN if either haplotype in a pair is NaN. 
                Otherwise, computes the mean while ignoring NaNs (e.g., 0|NaN -> 0, 1|NaN -> 1).
            is_weighted (bool, default=False): 
                If `True`, assigns individual weights from the `weight` column in `labels_file`. Otherwise, all individuals have equal weight of `1`.
            groups_to_remove (list of str, optional): 
                List with groups to exclude from analysis. Example: ['group1', 'group2'].
            min_percent_snps (float, default=4): 
                Minimum percentage of SNPs that must be known for an individual and of the ancestry of interest to be included in the analysis.
                All individuals with fewer percent of unmasked SNPs than this threshold will be excluded.
            group_snp_frequencies_only (bool, default=True):
                If True, mdPCA is performed exclusively on group-level SNP frequencies, ignoring individual-level data. This applies when `is_weighted` is 
                set to True and a `combination` column is provided in the `labels_file`,  meaning individuals are aggregated into groups based on their assigned 
                labels. If False, mdPCA is performed on individual-level SNP data alone or on both individual-level and group-level SNP frequencies when 
                `is_weighted` is True and a `combination` column is provided.
            save_masks (bool, default=False): 
                True if the masked matrices are to be saved in a `.npz` file, or False otherwise.
            load_masks (bool, default=False): 
                True if the masked matrices are to be loaded from a pre-existing `.npz` file specified by `masks_file`, or False otherwise.
            masks_file (str or pathlib.Path, default='masks.npz'): 
                Path to the `.npz` file used for saving/loading masked matrices.
            output_file (str or pathlib.Path, default='output.tsv'): 
                Path to the output `.tsv` file where mdPCA results are saved.
            covariance_matrix_file (str, optional): 
                Path to save the covariance matrix file in `.npy` format. If None, the covariance matrix is not saved. Default is None.
            n_components (int, default=2): 
                The number of principal components.
            rsid_or_chrompos (int, default=2): 
                Format indicator for SNP IDs in `self.__X_new_`. Use 1 for `rsID` format or 2 for `chromosome_position`.
            percent_vals_masked (float, default=0): 
                Percentage of values in the covariance matrix to be masked and then imputed. Only applicable if `method` is 
                `'cov_matrix_imputation'` or `'cov_matrix_imputation_ils'`.
        """
        self.__snpobj = snpobj
        self.__laiobj = laiobj
        self.__labels_file = labels_file
        self.__ancestry = self._define_ancestry(ancestry, laiobj.ancestry_map)
        self.__method = method
        self.__is_masked = is_masked
        self.__average_strands = average_strands
        self.__force_nan_incomplete_strands = force_nan_incomplete_strands
        self.__is_weighted = is_weighted
        self.__groups_to_remove = groups_to_remove
        self.__min_percent_snps = min_percent_snps
        self.__group_snp_frequencies_only = group_snp_frequencies_only
        self.__save_masks = save_masks
        self.__load_masks = load_masks
        self.__masks_file = masks_file
        self.__output_file = output_file
        self.__covariance_matrix_file = covariance_matrix_file
        self.__n_components = n_components
        self.__rsid_or_chrompos = rsid_or_chrompos
        self.__percent_vals_masked = percent_vals_masked
        self.__X_new_ = None  # Store transformed SNP data
        self.__haplotypes_ = None  # Store haplotypes of X_new_ (after filtering if min_percent_snps > 0)
        self.__samples_ = None  # Store samples of X_new_ (after filtering if min_percent_snps > 0)
        self.__variants_id_ = None  # Store variants ID (after filtering SNPs not in laiobj)

        # Fit and transform if a `snpobj`, `laiobj`, `labels_file`, and `ancestry` are provided
        if self.snpobj is not None and self.laiobj is not None and self.labels_file is not None and self.ancestry is not None:
            self.fit_transform(snpobj, laiobj, labels_file, ancestry)

    def __getitem__(self, key):
        """
        To access an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')

    def __setitem__(self, key, value):
        """
        To set an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            setattr(self, key, value)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')

    @property
    def method(self) -> str:
        """
        Retrieve `method`.
        
        Returns:
            **str:** The PCA method to use for dimensionality reduction.
        """
        return self.__method

    @method.setter
    def method(self, x: str) -> None:
        """
        Update `method`.
        """
        self.__method = x

    @property
    def snpobj(self) -> Optional['SNPObject']:
        """
        Retrieve `snpobj`.
        
        Returns:
            **SNPObject:** A SNPObject instance.
        """
        return self.__snpobj

    @snpobj.setter
    def snpobj(self, x: 'SNPObject') -> None:
        """
        Update `snpobj`.
        """
        self.__snpobj = x

    @property
    def laiobj(self) -> Optional['LocalAncestryObject']:
        """
        Retrieve `laiobj`.
        
        Returns:
            **LocalAncestryObject:** A LocalAncestryObject instance.
        """
        return self.__laiobj

    @laiobj.setter
    def laiobj(self, x: 'LocalAncestryObject') -> None:
        """
        Update `laiobj`.
        """
        self.__laiobj = x

    @property
    def labels_file(self) -> Optional[str]:
        """
        Retrieve `labels_file`.
        
        Returns:
            **str:** Path to the labels file in `.tsv` format.
        """
        return self.__labels_file

    @labels_file.setter
    def labels_file(self, x: str) -> None:
        """
        Update `labels_file`.
        """
        self.__labels_file = x

    @property
    def ancestry(self) -> Optional[str]:
        """
        Retrieve `ancestry`.
        
        Returns:
            **str:** Ancestry for which dimensionality reduction is to be performed. Ancestry counter starts at 0.
        """
        return self.__ancestry

    @ancestry.setter
    def ancestry(self, x: str) -> None:
        """
        Update `ancestry`.
        """
        self.__ancestry = x

    @property
    def is_masked(self) -> bool:
        """
        Retrieve `is_masked`.
        
        Returns:
            **bool:** True if an ancestry file is passed for ancestry-specific masking, or False otherwise.
        """
        return self.__is_masked

    @is_masked.setter
    def is_masked(self, x: bool) -> None:
        """
        Update `is_masked`.
        """
        self.__is_masked = x

    @property
    def average_strands(self) -> bool:
        """
        Retrieve `average_strands`.
        
        Returns:
            **bool:** True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
        """
        return self.__average_strands

    @average_strands.setter
    def average_strands(self, x: bool) -> None:
        """
        Update `average_strands`.
        """
        self.__average_strands = x

    @property
    def force_nan_incomplete_strands(self) -> bool:
        """
        Retrieve `force_nan_incomplete_strands`.
        
        Returns:
            **bool**: If `True`, sets the result to NaN if either haplotype in a pair is NaN.
                      Otherwise, computes the mean while ignoring NaNs (e.g., 0|NaN -> 0, 1|NaN -> 1).
        """
        return self.__force_nan_incomplete_strands

    @force_nan_incomplete_strands.setter
    def force_nan_incomplete_strands(self, x: bool) -> None:
        """
        Update `force_nan_incomplete_strands`.
        """
        self.__force_nan_incomplete_strands = x

    @property
    def is_weighted(self) -> bool:
        """
        Retrieve `is_weighted`.
        
        Returns:
            **bool:** True if weights are provided in the labels file, or False otherwise.
        """
        return self.__is_weighted

    @is_weighted.setter
    def is_weighted(self, x: bool) -> None:
        """
        Update `is_weighted`.
        """
        self.__is_weighted = x

    @property
    def groups_to_remove(self) -> Dict[int, List[str]]:
        """
        Retrieve `groups_to_remove`.
        
        Returns:
            **dict of int to list of str:** Dictionary specifying groups to exclude from analysis. Keys are array numbers, and values are 
                lists of groups to remove for each array. Example: `{1: ['group1', 'group2'], 2: [], 3: ['group3']}`.
        """
        return self.__groups_to_remove

    @groups_to_remove.setter
    def groups_to_remove(self, x: Dict[int, List[str]]) -> None:
        """
        Update `groups_to_remove`.
        """
        self.__groups_to_remove = x

    @property
    def min_percent_snps(self) -> float:
        """
        Retrieve `min_percent_snps`.
        
        Returns:
            **float:** 
                Minimum percentage of SNPs that must be known for an individual to be included in the analysis.
                All individuals with fewer percent of unmasked SNPs than this threshold will be excluded.
        """
        return self.__min_percent_snps

    @min_percent_snps.setter
    def min_percent_snps(self, x: float) -> None:
        """
        Update `min_percent_snps`.
        """
        self.__min_percent_snps = x

    @property
    def group_snp_frequencies_only(self) -> bool:
        """
        Retrieve `group_snp_frequencies_only`.
        
        Returns:
            **bool:** 
                If True, mdPCA is performed exclusively on group-level SNP frequencies, ignoring individual-level data. This applies 
                when `is_weighted` is set to True and a `combination` column is provided in the `labels_file`,  meaning individuals are 
                aggregated into groups based on their assigned labels. If False, mdPCA is performed on individual-level SNP data alone 
                or on both individual-level and group-level SNP frequencies when `is_weighted` is True and a `combination` column is provided.
        """
        return self.__group_snp_frequencies_only

    @group_snp_frequencies_only.setter
    def group_snp_frequencies_only(self, x: bool) -> None:
        """
        Update `group_snp_frequencies_only`.
        """
        self.__group_snp_frequencies_only = x

    @property
    def save_masks(self) -> bool:
        """
        Retrieve `save_masks`.
        
        Returns:
            **bool:** True if the masked matrices are to be saved in a `.npz` file, or False otherwise.
        """
        return self.__save_masks

    @save_masks.setter
    def save_masks(self, x: bool) -> None:
        """
        Update `save_masks`.
        """
        self.__save_masks = x

    @property
    def load_masks(self) -> bool:
        """
        Retrieve `load_masks`.
        
        Returns:
            **bool:** 
                True if the masked matrices are to be loaded from a pre-existing `.npz` file specified 
                by `masks_file`, or False otherwise.
        """
        return self.__load_masks

    @load_masks.setter
    def load_masks(self, x: bool) -> None:
        """
        Update `load_masks`.
        """
        self.__load_masks = x

    @property
    def masks_file(self) -> Union[str, pathlib.Path]:
        """
        Retrieve `masks_file`.
        
        Returns:
            **str or pathlib.Path:** Path to the `.npz` file used for saving/loading masked matrices.
        """
        return self.__masks_file

    @masks_file.setter
    def masks_file(self, x: Union[str, pathlib.Path]) -> None:
        """
        Update `masks_file`.
        """
        self.__masks_file = x

    @property
    def output_file(self) -> Union[str, pathlib.Path]:
        """
        Retrieve `output_file`.
        
        Returns:
            **str or pathlib.Path:** Path to the output `.tsv` file where mdPCA results are saved.
        """
        return self.__output_file

    @output_file.setter
    def output_file(self, x: Union[str, pathlib.Path]) -> None:
        """
        Update `output_file`.
        """
        self.__output_file = x

    @property
    def covariance_matrix_file(self) -> Optional[str]:
        """
        Retrieve `covariance_matrix_file`.
        
        Returns:
            **str:** Path to save the covariance matrix file in `.npy` format.
        """
        return self.__covariance_matrix_file
    
    @covariance_matrix_file.setter
    def covariance_matrix_file(self, x: Optional[str]) -> None:
        """
        Update `covariance_matrix_file`.
        """
        self.__covariance_matrix_file = x

    @property
    def n_components(self) -> int:
        """
        Retrieve `n_components`.
        
        Returns:
            **int:** The number of principal components.
        """
        return self.__n_components

    @n_components.setter
    def n_components(self, x: int) -> None:
        """
        Update `n_components`.
        """
        self.__n_components = x

    @property
    def rsid_or_chrompos(self) -> int:
        """
        Retrieve `rsid_or_chrompos`.
        
        Returns:
            **int:** Format indicator for SNP IDs in the SNP data. Use 1 for `rsID` format or 2 for `chromosome_position`.
        """
        return self.__rsid_or_chrompos

    @rsid_or_chrompos.setter
    def rsid_or_chrompos(self, x: int) -> None:
        """
        Update `rsid_or_chrompos`.
        """
        self.__rsid_or_chrompos = x

    @property
    def percent_vals_masked(self) -> float:
        """
        Retrieve `percent_vals_masked`.
        
        Returns:
            **float:** 
                Percentage of values in the covariance matrix to be masked and then imputed. Only applicable if `method` is 
                `'cov_matrix_imputation'` or `'cov_matrix_imputation_ils'`.
        """
        return self.__percent_vals_masked

    @percent_vals_masked.setter
    def percent_vals_masked(self, x: float) -> None:
        """
        Update `percent_vals_masked`.
        """
        self.__percent_vals_masked = x

    @property
    def X_new_(self) -> Optional[np.ndarray]:
        """
        Retrieve `X_new_`.

        Returns:
            **array of shape (n_samples, n_components):** 
                The transformed SNP data projected onto the `n_components` principal components.
                n_haplotypes_ is the number of haplotypes, potentially reduced if filtering is applied 
                (`min_percent_snps > 0`). For diploid individuals without filtering, the shape is 
                `(n_samples * 2, n_components)`.
        """
        return self.__X_new_

    @X_new_.setter
    def X_new_(self, x: np.ndarray) -> None:
        """
        Update `X_new_`.
        """
        self.__X_new_ = x

    @property
    def haplotypes_(self) -> Optional[List[str]]:
        """
        Retrieve `haplotypes_`.

        Returns:
            list of str:
                A list of unique haplotype identifiers.
        """
        if isinstance(self.__haplotypes_, np.ndarray):
            return self.__haplotypes_.ravel().tolist()  # Flatten and convert NumPy array to a list
        elif isinstance(self.__haplotypes_, list):
            if len(self.__haplotypes_) == 1 and isinstance(self.__haplotypes_[0], np.ndarray):
                return self.__haplotypes_[0].ravel().tolist()  # Handle list containing a single array
            return self.__haplotypes_  # Already a flat list
        elif self.__haplotypes_ is None:
            return None  # If no haplotypes are set
        else:
            raise TypeError("`haplotypes_` must be a list or a NumPy array.")

    @haplotypes_.setter
    def haplotypes_(self, x: Union[np.ndarray, List[str]]) -> None:
        """
        Update `haplotypes_`.
        """
        if isinstance(x, np.ndarray):
            self.__haplotypes_ = x.ravel().tolist()  # Flatten and convert to a list
        elif isinstance(x, list):
            if len(x) == 1 and isinstance(x[0], np.ndarray):  # Handle list containing a single array
                self.__haplotypes_ = x[0].ravel().tolist()
            else:
                self.__haplotypes_ = x  # Use directly if already a list
        else:
            raise TypeError("`x` must be a list or a NumPy array.")

    @property
    def samples_(self) -> Optional[List[str]]:
        """
        Retrieve `samples_`.

        Returns:
            list of str:
                A list of sample identifiers based on `haplotypes_` and `average_strands`.
        """
        haplotypes = self.haplotypes_
        if haplotypes is None:
            return None
        if self.__average_strands:
            return haplotypes
        else:
            return [x[:-2] for x in haplotypes]

    @property
    def variants_id_(self) -> Optional[np.ndarray]:
        """
        Retrieve `variants_id_`.

        Returns:
            **array of shape (n_snp,):** 
                An array containing unique identifiers (IDs) for each SNP,
                potentially reduced if there are SNPs not present in the `laiobj`.
                The format will depend on `rsid_or_chrompos`.
        """
        return self.__variants_id_

    @variants_id_.setter
    def variants_id_(self, x: np.ndarray) -> None:
        """
        Update `variants_id_`.
        """
        self.__variants_id_ = x

    @property
    def n_haplotypes(self) -> Optional[int]:
        """
        Retrieve `n_haplotypes`.

        Returns:
            **int:**
                The total number of haplotypes, potentially reduced if filtering is applied 
                (`min_percent_snps > 0`).
        """
        return len(self.haplotypes_)

    @property
    def n_samples(self) -> Optional[int]:
        """
        Retrieve `n_samples`.

        Returns:
            **int:**
                The total number of samples, potentially reduced if filtering is applied 
                (`min_percent_snps > 0`).
        """
        return len(np.unique(self.samples_))

    def copy(self) -> 'mdPCA':
        """
        Create and return a copy of `self`.

        Returns:
            **mdPCA:** 
                A new instance of the current object.
        """
        return copy.copy(self)

    @staticmethod
    def _define_ancestry(ancestry, ancestry_map):
        """
        Determine the ancestry index based on different input types.

        Args:
            ancestry (int or str): The ancestry input, which can be:
                - An integer (e.g., 0, 1, 2).
                - A string representation of an integer (e.g., '0', '1').
                - A string matching one of the ancestry map values (e.g., 'Africa').
            ancestry_map (dict): A dictionary mapping ancestry indices (as strings) to ancestry names.

        Returns:
            int: The corresponding ancestry index.
        """
        if isinstance(ancestry, int):  
            return ancestry  
        elif isinstance(ancestry, str) and ancestry.isdigit():  
            return int(ancestry)  
        elif ancestry in ancestry_map.values():  
            return int(next(key for key, value in ancestry_map.items() if value == ancestry))  
        else:  
            raise ValueError(f"Invalid ancestry input: {ancestry}")

    def _load_mask_file(self):
        """
        Load previously saved masked genotype data from an `.npz` file.
        """
        mask_files = np.load(self.masks_file, allow_pickle=True)
        mask = mask_files['mask']
        rs_ID_list = mask_files['rs_ID_list']
        ind_ID_list = mask_files['ind_ID_list']
        labels = mask_files['labels']
        weights = mask_files['weights']
        return mask, rs_ID_list, ind_ID_list, labels, weights

    @staticmethod
    def _compute_strength_vector(X):
        strength_vector = np.sum(~np.isnan(X), axis=1) / X.shape[1]
        return strength_vector

    @staticmethod
    def _compute_strength_matrix(X):
        notmask = (~np.isnan(X)).astype(np.float32)
        strength_matrix = np.dot(notmask, notmask.T)
        strength_matrix /= X.shape[1]
        return strength_matrix

    def _cov(self, x):
        ddof = 1
        x = np.ma.array(x, ndmin=2, copy=True, dtype=np.float32)
        xmask = np.ma.getmaskarray(x)
        xnotmask = np.logical_not(xmask).astype(np.float32)
        fact = np.dot(xnotmask, xnotmask.T) * 1. - ddof
        del xnotmask
        gc.collect()
        result = (np.ma.dot(x, x.T, strict=False) / fact).squeeze()
        x = x.data
        strength_vec = self._compute_strength_vector(x)
        strength_mat = self._compute_strength_matrix(x)
        return result.data, strength_vec, strength_mat

    @staticmethod
    def _demean(S, w):
        w_sum = np.matmul(w, S)
        w_rowsum = w_sum.reshape((1, S.shape[0]))
        w_colsum = w_sum.reshape((S.shape[0], 1))
        w_totalsum = np.dot(w_sum, w)
        S -= (w_rowsum + w_colsum) - w_totalsum
        return S

    @staticmethod
    def _iterative_svd_matrix_completion(cov):
        num_masked = np.isnan(cov).sum()
        if num_masked > 0:
            start_time = time.time()
            start_rank = 1
            end_rank = 10
            rank = 5
            choose_best = True
            num_cores = 5
            cov_complete = IterativeSVD(start_rank=start_rank, end_rank=end_rank, rank=rank, choose_best=choose_best, num_cores=num_cores).fit_transform(cov)
            logging.info("Iterative SVD --- %s seconds ---" % (time.time() - start_time))
        else:
            cov_complete = cov
        return cov_complete

    def _weighted_cov_pca(self, X_incomplete, weights, covariance_matrix_file, n_components):
        start_time = time.time()
        X_incomplete = np.ma.array(X_incomplete, mask=np.isnan(X_incomplete))
        S, _, _ = self._cov(X_incomplete)
        logging.info("Covariance Matrix --- %s seconds ---" % (time.time() - start_time))
        weights_normalized = weights / weights.sum()
        S = self._demean(S, weights_normalized)
        W = np.diag(weights)
        WSW = np.matmul(np.matmul(np.sqrt(W), S), np.sqrt(W))
        if covariance_matrix_file:
            np.save(covariance_matrix_file, S.data)
        svd = TruncatedSVD(n_components, algorithm="arpack")
        svd.fit(WSW)
        U = np.matmul(svd.components_, np.linalg.inv(np.sqrt(W))).T
        D = np.diag(np.sqrt(svd.singular_values_))
        T = np.matmul(U, D)
        total_var = np.trace(S)
        for i in range(n_components):
            pc_percentvar = 100 * svd.singular_values_[i] / total_var
            logging.info("Percent variance explained by the principal component %d: %s", i + 1, pc_percentvar)
        return T

    @staticmethod
    def _create_validation_mask(X_incomplete, percent_inds):
        np.random.seed(1)
        masked_rows = np.isnan(X_incomplete).any(axis=1)
        masked_inds = np.flatnonzero(masked_rows)
        X_masked = X_incomplete[masked_rows] 
        percent_masked = 100 * np.isnan(X_masked).sum() / (X_masked.shape[0] * X_masked.shape[1])
        unmasked_rows = ~masked_rows
        X_unmasked = X_incomplete[unmasked_rows]
        masked_rows = np.random.choice(range(X_unmasked.shape[0]), size=int(X_unmasked.shape[0] * percent_inds / 100), replace=False)
        X_masked_rows = X_unmasked[masked_rows,:]
        mask = np.zeros(X_masked_rows.shape[0] * X_masked_rows.shape[1], dtype=np.int8)
        mask[:int(X_masked_rows.shape[0] * X_masked_rows.shape[1] * percent_masked / 100)] = 1
        np.random.shuffle(mask)
        mask = mask.astype(bool)
        mask = mask.reshape(X_masked_rows.shape)
        X_masked_rows[mask] = np.nan
        X_unmasked[masked_rows] = X_masked_rows
        X_incomplete[unmasked_rows] = X_unmasked
        masked_rows_new = np.isnan(X_incomplete).any(axis=1)
        masked_inds_new = np.flatnonzero(masked_rows_new)
        masked_inds_val = sorted(list(set(masked_inds_new) - set(masked_inds)))
        return X_incomplete, masked_inds_val

    @staticmethod    
    def _create_cov_mask(cov, strength_mat, percent_vals_masked):
        cov_flattened = cov.reshape(-1).copy()
        num_vals_masked = int(percent_vals_masked * len(cov_flattened) / 100)
        pos_masked = strength_mat.reshape(-1).argsort()[:num_vals_masked]
        cov_flattened[pos_masked] = np.nan
        cov_masked = cov_flattened.reshape(cov.shape)
        return cov_masked

    def _regularized_optimization_ils(self, X_incomplete, covariance_matrix_file, n_components):
        def run_cov_matrix_regularized_optimization_ils(X_incomplete, covariance_matrix_file):
            start_time = time.time()
            X_incomplete = np.ma.array(X_incomplete, mask=np.isnan(X_incomplete))
            has_missing = np.isnan(X_incomplete.data).sum() > 0
            S, _, strength_mat = self._cov(X_incomplete)
            logging.info("Covariance Matrix --- %s seconds ---" % (time.time() - start_time))
            S = self._demean(S, np.ones(S.shape[0]) / S.shape[0])
            if has_missing: 
                logging.info("Starting matrix completion. This will take a few minutes...")
                start_time = time.time()
                percent_inds_val = 20 # Percent of unmasked individuals to be masked for cross-validation 
                X_incomplete, masked_inds_val = self._create_validation_mask(X_incomplete.data, percent_inds_val) # masked_inds_val is the list of indices of the individuals masked for validation
                X_incomplete = np.ma.array(X_incomplete, mask=np.isnan(X_incomplete))
                S_prime, _, w_mat_prime = self._cov(X_incomplete)
                del X_incomplete
                gc.collect()
                S_prime = self._demean(S_prime, np.ones(S_prime.shape[0]) / S_prime.shape[0])
                S_robust, lam = self.matrix_completion(S, strength_mat, S_prime, w_mat_prime, lams=None, method="NN", 
                                                cv_inds=masked_inds_val)
                logging.info(f"Covariance Matrix --- %{time.time() - start_time:.2}s seconds ---")
            else:
                S_robust = S.copy()
            if covariance_matrix_file:
                np.save(covariance_matrix_file, S.data)
                if has_missing:
                    base, ext = os.path.splitext(covariance_matrix_file)
                    np.save(f"{base}_completed_{lam}{ext}", S_robust.data)
            return S, S_robust

        def _compute_projection_regularized_optimization_ils(X_incomplete, S, S_robust, n_components):
            _, _, strength_mat = self._cov(np.ma.array(X_incomplete, mask=np.isnan(X_incomplete)))
            nonmissing = np.flatnonzero(np.isnan(X_incomplete).sum(axis=1) == 0)
            missing = np.flatnonzero(np.isnan(X_incomplete).sum(axis=1) > 0)
            missing_sorted = sorted(missing, key=lambda x: (~np.isnan(X_incomplete[x])).sum(), reverse=True)
            svd = TruncatedSVD(n_components, algorithm="arpack")
            svd.fit(S_robust)
            U = svd.components_.T
            U_nonmissing = U[nonmissing]
            D_nonmissing = np.diag(np.sqrt(svd.singular_values_))
            T = np.zeros((X_incomplete.shape[0], n_components))
            T_filled = np.matmul(U_nonmissing, D_nonmissing)
            i_filled = nonmissing
            T[nonmissing] = T_filled
            for i in missing_sorted:
                S_i = S[i][i_filled]
                W_i = np.diag(strength_mat[i][i_filled])
                A = np.matmul(W_i, T_filled)
                b = np.matmul(W_i, S_i)
                t = np.linalg.lstsq(A, b, rcond=None)[0]
                T[i] = t
                i_filled = np.append(i_filled, i)
                T_filled = np.append(T_filled, [t], axis=0)
            total_var = np.trace(S)
            for i in range(n_components):
                pc_percentvar = 100 * svd.singular_values_[i] / total_var
                logging.info("Percent variance explained by the principal component %d: %s", i + 1, pc_percentvar)
            return T
        
        S, S_robust = run_cov_matrix_regularized_optimization_ils(X_incomplete, covariance_matrix_file)
        T = _compute_projection_regularized_optimization_ils(X_incomplete, S, S_robust, n_components)
        return T

    def _cov_matrix_imputation(self, X_incomplete, percent_vals_masked, covariance_matrix_file, n_components):
        start_time = time.time()
        X_incomplete = np.ma.array(X_incomplete, mask=np.isnan(X_incomplete))
        S, strength_vec, strength_mat = self._cov(X_incomplete)
        logging.info("Covariance Matrix --- %s seconds ---" % (time.time() - start_time))
        S_masked = self._create_cov_mask(S, strength_mat, percent_vals_masked)
        S = self._iterative_svd_matrix_completion(S_masked)
        weights = np.ones(len(strength_vec))
        weights_normalized = weights / weights.sum()
        S = self._demean(S, weights_normalized)
        W = np.diag(weights)
        WSW = np.matmul(np.matmul(np.sqrt(W), S), np.sqrt(W))
        if covariance_matrix_file:
            np.save(covariance_matrix_file, S.data)
        svd = TruncatedSVD(n_components, algorithm="arpack")
        svd.fit(WSW)
        U = np.matmul(svd.components_, np.linalg.inv(np.sqrt(W))).T
        D = np.diag(np.sqrt(svd.singular_values_))
        T = np.matmul(U, D)
        total_var = np.trace(S)
        for i in range(n_components):
            pc_percentvar = 100 * svd.singular_values_[i] / total_var
            logging.info("Percent variance explained by the principal component %d: %s", i + 1, pc_percentvar)
        return T

    def _cov_matrix_imputation_ils(self, X_incomplete, percent_vals_masked, covariance_matrix_file, n_components):
        
        def run_cov_matrix_cov_matrix_imputation_ils(X_incomplete, covariance_matrix_file):
            start_time = time.time()
            X_incomplete = np.ma.array(X_incomplete, mask=np.isnan(X_incomplete))
            S, strength_vec, strength_mat = self._cov(X_incomplete)
            logging.info("Covariance Matrix --- %s seconds ---" % (time.time() - start_time))
            S_masked = self._create_cov_mask(S, strength_mat, percent_vals_masked)
            S = self._iterative_svd_matrix_completion(S_masked)
            weights = np.ones(len(strength_vec))
            weights_normalized = weights / weights.sum()
            S = self._demean(S, weights_normalized)
            W = np.diag(weights)
            if covariance_matrix_file:
                np.save(covariance_matrix_file, S.data)
            return S, W

        def compute_projection_cov_matrix_imputation_ils(X_incomplete, S, W, n_components):
            S, _, strength_mat = self._cov(np.ma.array(X_incomplete, mask=np.isnan(X_incomplete)))
            S = self._demean(S, np.ones(S.shape[0]) / S.shape[0])
            nonmissing = np.flatnonzero(np.isnan(X_incomplete).sum(axis=1) == 0)
            missing = np.flatnonzero(np.isnan(X_incomplete).sum(axis=1) > 0)
            missing_sorted = sorted(missing, key=lambda x: (~np.isnan(X_incomplete[x])).sum(), reverse=True)
            svd = TruncatedSVD(n_components, algorithm="arpack")
            WSW = np.matmul(np.matmul(np.sqrt(W), S), np.sqrt(W))
            svd.fit(WSW)
            U = np.matmul(svd.components_, np.linalg.inv(np.sqrt(W))).T
            U_nonmissing = U[nonmissing]
            D_nonmissing = np.diag(np.sqrt(svd.singular_values_))
            T = np.zeros((X_incomplete.shape[0], n_components))
            T_filled = np.matmul(U_nonmissing, D_nonmissing)
            i_filled = nonmissing
            T[nonmissing] = T_filled
            for i in missing_sorted:
                S_i = S[i][i_filled]
                W_i = np.diag(strength_mat[i][i_filled])
                A = np.matmul(W_i, T_filled)
                b = np.matmul(W_i, S_i)
                t = np.linalg.lstsq(A, b, rcond=None)[0]
                T[i] = t
                i_filled = np.append(i_filled, i)
                T_filled = np.append(T_filled, [t], axis=0)
            total_var = np.trace(S)
            for i in range(n_components):
                pc_percentvar = 100 * svd.singular_values_[i] / total_var
                logging.info("Percent variance explained by the principal component %d: %s", i + 1, pc_percentvar)
            return T
        
        S, W = run_cov_matrix_cov_matrix_imputation_ils(X_incomplete, covariance_matrix_file)
        T = compute_projection_cov_matrix_imputation_ils(X_incomplete, S, W, n_components)
        return T

    def _nonmissing_pca_ils(self, X_incomplete, covariance_matrix_file, n_components):
    
        def run_cov_matrix_nonmissing_pca_ils(X_incomplete, covariance_matrix_file):
            start_time = time.time()
            nonmissing = np.flatnonzero(np.isnan(X_incomplete).sum(axis=1) == 0)
            X_incomplete_nonmissing = X_incomplete[nonmissing]
            X_incomplete_nonmissing = np.ma.array(X_incomplete_nonmissing, mask=np.isnan(X_incomplete_nonmissing))
            S_nonmissing, _, _ = self._cov(X_incomplete_nonmissing)
            logging.info("Covariance Matrix --- %s seconds ---" % (time.time() - start_time))
            S_nonmissing = self._demean(S_nonmissing, np.ones(S_nonmissing.shape[0]) / S_nonmissing.shape[0])    
            if covariance_matrix_file:
                np.save(covariance_matrix_file, S_nonmissing.data)
            return S_nonmissing
        
        def compute_projection_nonmissing_pca_ils(X_incomplete, S_nonmissing, n_components):
            S, _, strength_mat = self._cov(np.ma.array(X_incomplete, mask=np.isnan(X_incomplete)))
            S = self._demean(S, np.ones(S.shape[0]) / S.shape[0])
            nonmissing = np.flatnonzero(np.isnan(X_incomplete).sum(axis=1) == 0)
            missing = np.flatnonzero(np.isnan(X_incomplete).sum(axis=1) > 0)
            missing_sorted = sorted(missing, key=lambda x: (~np.isnan(X_incomplete[x])).sum(), reverse=True)
            svd = TruncatedSVD(n_components, algorithm="arpack")
            svd.fit(S_nonmissing)
            U_nonmissing = svd.components_.T
            D_nonmissing = np.diag(np.sqrt(svd.singular_values_))
            T = np.zeros((X_incomplete.shape[0], n_components))
            T_filled = np.matmul(U_nonmissing, D_nonmissing)
            i_filled = nonmissing
            T[nonmissing] = T_filled
            for i in missing_sorted:
                S_i = S[i][i_filled]
                W_i = np.diag(strength_mat[i][i_filled])
                A = np.matmul(W_i, T_filled)
                b = np.matmul(W_i, S_i)
                t = np.linalg.lstsq(A, b, rcond=None)[0]
                T[i] = t
                i_filled = np.append(i_filled, i)
                T_filled = np.append(T_filled, [t], axis=0)
            total_var = np.trace(S)
            for i in range(n_components):
                pc_percentvar = 100 * svd.singular_values_[i] / total_var
                logging.info("Percent variance explained by the principal component %d: %s", i + 1, pc_percentvar)
            return T
        
        S_nonmissing = run_cov_matrix_nonmissing_pca_ils(X_incomplete, covariance_matrix_file)
        T = compute_projection_nonmissing_pca_ils(X_incomplete, S_nonmissing, n_components)
        return T

    def _run_cov_matrix(self, X_incomplete, weights):
        """
        Runs the specified PCA method on the incomplete data.
        """
        if self.method == "weighted_cov_pca":
            return self._weighted_cov_pca(X_incomplete, weights, self.covariance_matrix_file, self.n_components)
        elif self.method == "regularized_optimization_ils":
            return self._regularized_optimization_ils(X_incomplete, self.covariance_matrix_file, self.n_components)
        elif self.method == "cov_matrix_imputation":
            return self._cov_matrix_imputation(X_incomplete, self.percent_vals_masked, self.covariance_matrix_file, self.n_components)
        elif self.method == "cov_matrix_imputation_ils":
            return self._cov_matrix_imputation_ils(X_incomplete, self.percent_vals_masked, self.covariance_matrix_file, self.n_components)
        elif self.method == "nonmissing_pca_ils":
            return self._nonmissing_pca_ils(X_incomplete, self.covariance_matrix_file, self.n_components)
        else:
            raise ValueError(f"Unsupported method: {self.method}")

    def fit_transform(
            self,
            snpobj: Optional['SNPObject'] = None, 
            laiobj: Optional['LocalAncestryObject'] = None,
            labels_file: Optional[str] = None,
            ancestry: Optional[str] = None,
            average_strands: Optional[bool] = None
        ) -> np.ndarray:
        """
        Fit the model to the SNP data stored in the provided `snpobj` and apply the dimensionality reduction on the same SNP data.
        
        This method starts by loading or updating SNP and ancestry data. Then, it manages missing values by applying 
        masks based on ancestry, either by loading a pre-existing mask or generating new ones. After processing these 
        masks to produce an incomplete SNP data matrix, it applies the chosen mdPCA method to reduce dimensionality 
        while handling missing data as specified.

        Args:
            snpobj (SNPObject, optional): 
                A SNPObject instance.
            laiobj (LAIObject, optional): 
                A LAIObject instance.
            labels_file (str, optional): 
                Path to the labels file in .tsv format. The first column, `indID`, contains the individual identifiers, and the second 
                column, `label`, specifies the groups for all individuals. If `is_weighted=True`, a `weight` column with individual 
                weights is required. Optionally, `combination` and `combination_weight` columns can specify sets of individuals to be 
                combined into groups, with respective weights.
            ancestry (str, optional): 
                Ancestry for which dimensionality reduction is to be performed. Ancestry counter starts at 0.
            average_strands (bool, optional): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
                If None, defaults to `self.average_strands`.

        Returns:
            **array of shape (n_samples, n_components):** 
                The transformed SNP data projected onto the `n_components` principal components, stored in `self.X_new_`.
        """
        if snpobj is None:
            snpobj = self.snpobj
        if laiobj is None:
            laiobj = self.laiobj
        if labels_file is None:
            labels_file = self.labels_file
        if ancestry is None:
            ancestry = self.ancestry
        if average_strands is None:
            average_strands = self.average_strands
        
        if self.load_masks:
            # Load precomputed ancestry-based masked genotype matrix, SNP identifiers, haplotype identifiers, and weights
            mask, variants_id, haplotypes, _, weights = self._load_mask_file()
        else:
            # Process genotype data with optional ancestry-based masking and return the corresponding SNP and individual identifiers
            mask, variants_id, haplotypes = process_calldata_gt(
                self.snpobj,
                self.laiobj,
                self.ancestry,
                self.average_strands,
                self.force_nan_incomplete_strands,
                self.is_masked,
                self.rsid_or_chrompos
            )

            # Process individual genomic labels and weights, aligning them with a masked genotype matrix by 
            # filtering out low-coverage individuals, reordering data to match the matrix structure, and 
            # handling group-based adjustments
            mask, haplotypes, _, weights = process_labels_weights(
                self.labels_file,
                mask,
                variants_id,
                haplotypes,
                self.average_strands,
                self.ancestry,
                self.min_percent_snps,
                self.group_snp_frequencies_only,
                self.groups_to_remove,
                self.is_weighted,
                self.save_masks,
                self.masks_file
            )

        # Call run_cov_matrix with the specified method
        self.X_new_ = self._run_cov_matrix(
            mask[self.ancestry].T,
            weights
        )

        self.haplotypes_ = haplotypes
        self.variants_id_ = variants_id

        return self.X_new_



import numpy as np

from snputils import VCFReader, BEDReader, PGENReader


def test_vcf_summed_strands(snpobj_vcf, data_path):
    snpobj_vcf_summed_strands = VCFReader(data_path + "/vcf/subset.vcf").read(sum_strands=True)
    assert np.array_equal(snpobj_vcf.calldata_gt.sum(axis=2, dtype=np.int8), snpobj_vcf_summed_strands.calldata_gt)


def test_bed_summed_strands(snpobj_bed, data_path):
    snpojb_bed_summed_strands = BEDReader(data_path + "/bed/subset").read(sum_strands=True)
    assert np.array_equal(snpobj_bed.calldata_gt.sum(axis=2, dtype=np.int8), snpojb_bed_summed_strands.calldata_gt)


def test_pgen_summed_strands(snpobj_pgen, data_path):
    snpobj_pgen_summed_strands = PGENReader(data_path + "/pgen/subset").read(sum_strands=True)
    assert np.array_equal(snpobj_pgen.calldata_gt.sum(axis=2, dtype=np.int8), snpobj_pgen_summed_strands.calldata_gt)
